##############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################

"""Configuration variables for controlling specific things in VisTrails."""
from __future__ import division

import argparse
import ast
import copy
import itertools
import re
import shlex
import sys
import weakref

import os

from vistrails.core import system
from vistrails.core.utils import Ref, append_to_dict_of_lists
from vistrails.db.domain import DBConfiguration, DBConfigKey, DBConfigStr, \
    DBConfigInt, DBConfigFloat, DBConfigBool

##############################################################################

_docs = {}
_simple_docs = {}
_usage_args = set()

_simple_documentation = """
autoConnect: Automatically connect dragged in modules
autoSave: Automatically save backup vistrails every two minutes
batch: Run in batch mode instead of interactive mode
cache: Cache previous results so they may be used in future computations
dataDir: Default data directory
db: The name for the database to load the vistrail from
dbDefault: Save vistrails in a database by default
debugLevel: How much information should VisTrails log
defaultFileType: Default file type/extension for vistrails (.vt or .xml)
detachHistoryView: Show the version tree in a separate window
dotVistrails: User configuration directory
enablePackagesSilently: Automatically enable packages when needed
errorLog: Write errors to a log file
execute: Execute any specified workflows
executionLog: Track execution provenance when running workflows
fileDir: Default vistrail directory
fixedSpreadsheetCells: Draw spreadsheet cells at a fixed size
handlerDontAsk: Do not ask about extension handling at startup
hideUpgrades: Don't show upgrade nodes in the version tree
host: The hostname for the database to load the vistrail from
installBundles: Install missing Python dependencies
installBundlesWithPip: Use pip to install missing Python dependencies
isInServerMode: Indicates whether VisTrails is being run as a server
jobAutorun: Run jobs automatically when they finish
jobCheckInterval: How often to check for jobs (in seconds)
jobList: List running workflows
jobInfo: List jobs in running workflow
loadPackages: Whether to load the packages enabled in the configuration file
logDir: Log files directory
maxRecentVistrails: Number of recent vistrails
maximizeWindows: VisTrails windows should be maximized
migrateTags: Move tags to upgraded versions
multiHeads: Use multiple screens for VisTrails windows
multithread: Server will start a thread for each request
packageDir: System packages directory
parameterExploration: Run parameter exploration instead of workflow
parameters: List of parameters to use when running workflow
port: The port for the database to load the vistrail from
repositoryHTTPURL: Remote package repository URL
repositoryLocalPath: Local package repository directory
rootDirectory: Directory that contains the VisTrails source code
rpcConfig: Config file for server connection options
rpcInstances: Number of other instances that vistrails should start
rpcLogFile: Log file for XML RPC server
rpcPort: Port where this xml rpc server will work
rpcServer: Hostname or ip address where this xml rpc server will work
shell.fontFace: Console Font
shell.fontSize: Console Font Size
showConnectionErrors: Show error when input value doesn't match type during execution
showDebugPopups: Always bring debug messages to the front
showInlineParameterWidgets: Show editable parameters inside modules
showScrollbars: Show scrollbars on the version tree and workflow canvases
showSplash: Show VisTrails splash screen during startup
showSpreadsheetOnly: Hides the VisTrails main window
showVariantErrors: Show error when variant input value doesn't match type during execution
showWindow: Show the main window
singleInstance: Do not allow more than one instance of VisTrails to run at once
spreadsheetDumpCells: Defines the location for generated cells
spreadsheetDumpPDF: Whether the spreadsheet should dump images in PDF format
staticRegistry: XML registry file
stopOnError: Stop all workflow execution immediately after first error
subworkflowsDir: Local subworkflows directory
temporaryDir: Temporary files directory
thumbs.autoSave: Save thumbnails of visual results
thumbs.cacheDir: Thumbnail cache directory
thumbs.cacheSize: Thumbnail cache size (MB)
thumbs.mouseHover: Show thumbnails when mouse is hovering above a version
thumbs.tagsOnly: Store thumbnails only for tagged versions
upgradeDelay: Persist upgrade only after other changes
upgradeModuleFailPrompt: Alert when a subworkflow upgrade fails
upgrades: Attempt to automatically upgrade old workflows
useMacBrushedMetalStyle: Use a brushed metal interface (MacOS X only)
user: The username for the database to the load vistrail from
userPackageDir: Local packages directory
viewOnLoad: Whether to show pipeline or history view when opening vistrail
webRepositoryURL: Web repository URL
webRepositoryUser: Web repository username
withVersionTree: Output the version tree as an image
withWorkflow: Output the workflow graph as an image
"""

_documentation = """
autoConnect: Boolean

    Try to automatically connect a newly dragged in module to the rest
    of the workflow.

autoSave: Boolean

    Automatically save vistrails to allow recovery from crashes, etc.

batch: Boolean

    Run vistrails in batch mode instead of interactive mode.

cache: Boolean

    Cache previous results so they may be used in future computations.

dataDir: Path

    The location that VisTrails uses as a default directory for data.

db: String

    The name for the database to load the vistrail from.

dbDefault: Boolean

    Use a database as the default storage location for vistrails entities.

debugLevel: Integer

    How much information VisTrails should alert the user about (0:
    Critical errors only, 1: Critical errors and warnings, 2: Critical
    errors, warnings, and log messages).

defaultFileType: String

    Defaults to .vt but could be .xml.

detachHistoryView: Boolean

    Show the version tree in a separate window.

dotVistrails: Path

    The location to look for VisTrails user configurations and
    storage. Defaults to ~/.vistrails.

enablePackagesSilently: Boolean

    Do not prompt the user to enable packages, just do so
    automatically.

errorLog: Boolean

    Write errors to a log file.

execute: Boolean

    Execute any specified workflows.

executionLog: Boolean

    Track execution provenance when running workflows.

fileDir: Path

    The location that VisTrails uses as a default directory for
    specifying files.

fixedSpreadsheetCells: Boolean

    Draw spreadsheet cells at a fixed size (for testing).

handlerDontAsk: Boolean

    Do not ask about extension handling at startup (Linux only).

hideUpgrades: Boolean

    Don't show the "upgrade" nodes in the version tree.

host: URL

    The hostname for the database to load the vistrail from.

installBundles: Boolean

    Automatically try to install missing Python dependencies.

installBundlesWithPip: Boolean

    Whether to try to use pip to install Python dependencies or use
    distribution support.

isInServerMode: Boolean

    Indicates whether VisTrails is being run as a server.

jobAutorun: Boolean

    Run jobs automatically when they finish.

jobCheckInterval: Integer:

    How often to check for jobs (in seconds, default=600).

jobList: Boolean

    List running workflows.

jobInfo: Boolean

    List jobs in running workflow

loadPackages: Boolean

    Whether to load the packages enabled in the configuration file

logDir: Path

    The path that indicates where log files should be stored.

logger: ConfigurationObject

    *Deprecated*

maximizeWindows: Boolean

    Whether the VisTrails windows should take up the entire screen space.

maxMemory: Integer

    *Deprecated*

maxRecentVistrails: Integer

    How many recently opened vistrails should be stored for "Open
    Recent" access.

migrateTags: Boolean

    Whether or not the tag on a workflow that was upgraded should be
    moved to point to the upgraded version.

minMemory: Integer

    *Deprecated*

multiHeads: Boolean

    Whether or not to use multiple screens for VisTrails windows.

multithread: Boolean

    Server will start a thread for each request.

OLDupgradeDelay: Boolean
    If True, will only persist the upgrade when a user makes a
    modification to or executes the workflow. Otherwise, the upgrade
    will be automatically added to the version tree when a user views
    an upgraded workflow.

outputDefaultSettings: ConfigurationObject

    One or more comma-separated key=value parameters.

outputSettings: ConfigurationObject

    One or more comma-separated key=value parameters.

packageDir: Path

    The directory to look for VisTrails core packages (use
    userPackageDir for user-defined packages).

parameterExploration: Boolean

    Open and execute parameter exploration specified by the
    version argument after the .vt file.

parameters: String

    List of parameters to use when running workflow.

port: Integer

    The port for the database to load the vistrail from.

pythonPrompt: Boolean

    *Deprecated*

recentVistrailList: String

    Storage for recent vistrails. Users should not edit.

repositoryHTTPURL: URL

    URL used to locate packages available to be installed.

repositoryLocalPath: Path

    Path used to locate packages available to be installed.

reviewMode: Boolean

    *Deprecated* Used to interactively export a pipeline.

rootDirectory: Path

    Directory that contains the VisTrails source code.

rpcConfig: String

    Config file for server connection options.

rpcInstances: Integer

    Number of other instances that vistrails should start.

rpcLogFile: String

    Log file for XML RPC server.

rpcPort: Integer

    Port where this xml rpc server will work.

rpcServer: URL

    Hostname or ip address where this xml rpc server will work.

runningJobsList: String

    Storage for recent vistrails; users should not edit.

shell: ConfigurationObject

    Settings for the appearance of the VisTrails console.

shell.fontFace: String

    The font to be used for the VisTrails console.

shell.fontSize: Integer

    The font size used for the VisTrails console.

showConnectionErrors: Boolean

    Alert the user if the value along a connection doesn't match
    connection types.

showDebugPopups: Boolean

    Always show the debug popups or only if there is a modal widget.

showMovies: Boolean

    *Deprecated* Set automatic movie creation on the spreadsheet.

showScrollbars: Boolean

    Whether VisTrails should show scrollbars on the version tree and workflow
    canvases.

showSplash: Boolean

    Whether the VisTrails splash screen should be shown on startup.

showSpreadsheetOnly: Boolean

    Whether the VisTrails main window should be hidden.

showVariantErrors: Boolean

    Alert the user if the value along a connection coming from a
    Variant output doesn't match the input port.

showWindow: Boolean

    Show the main VisTrails window.

singleInstance: Boolean

    Whether or not VisTrails should only allow one instance to be
    running.

spreadsheetDumpCells: Path

    If specified, defines the location for parameter exploration to
    store the cells it generates.

spreadsheetDumpPDF: Boolean

    Whether the spreadsheet should dump images in PDF format.

staticRegistry: Path

    If specified, VisTrails uses an XML file defining the VisTrails
    module registry to load modules instead of from the packages directly.

stopOnError: Boolean

    Whether or not VisTrails stops executing the rest of the workflow
    if it encounters an error in one module.

subworkflowsDir: Path

    The location where a user's local subworkflows are stored.

temporaryDir: Path

    The directory to use for temporary files generated by VisTrails.

thumbs: ConfiguationObject

    Settings for generating and saving thumbnail images.

thumbs.autoSave: Boolean

    Whether to save thumbnails of results when executing VisTrails.

thumbs.cacheDir: Path

    The directory to be used to cache thumbnails.

thumbs.cacheSize: Integer

    The size (in MB) of the thumbnail cache.

thumbs.mouseHover: Boolean

    Whether to show thumbnails when hovering over a version in the
    version tree.

thumbs.tagsOnly: Boolean

    If True, only stores thumbnails for tagged versions. Otherwise,
    stores thumbnails for all versions.

upgradeDelay: Boolean

    Persist upgrade only after other changes.

upgrades: Boolean

    Whether to upgrade old workflows so they work with newer packages.

upgradeModuleFailPrompt: Boolean

    Whether to alert the user when an upgrade may fail when upgrading
    a subworkflow.

useMacBrushedMetalStyle: Boolean

    Whether should use a brushed metal interface (MacOS X only).

user: String

    The username for the database to load the vistrail from.

userPackageDir: Boolean

    The location for user-installed packages (defaults to
    ~/.vistrails/userpackages).

viewOnLoad: String

    Whether to show pipeline or history view when opening vistrail
    Can be either appropriate/pipeline/history

webRepositoryURL: URL

    The URL of the web repository that should be attached to VisTrails
    (e.g. www.crowdlabs.org).

webRepositoryUser: String

    The default username for logging into a VisTrails web repository
    like crowdLabs.

withVersionTree: Boolean

    Output the version tree as an image.

withWorkflow: Boolean

    Output the workflow graph as an image.

showInlineParameterWidgets: Boolean

    Show editable parameters inside modules

"""

class ConfigType(object):
    NORMAL = 0
    SHOW_HIDE = 1
    ON_OFF = 2
    COMMAND_LINE = 3
    COMMAND_LINE_FLAG = 4
    INTERNAL = 5
    STORAGE = 6
    PACKAGE = 7
    SUBOBJECT = 8
    INTERNAL_SUBOBJECT = 9

class ConfigString(object):
    @classmethod
    def from_string(cls, val):
        if not val:
            return None
        return val

    @classmethod
    def to_string(cls, val):
        if val is None:
            return ""
        return val

class ConfigPath(ConfigString):
    pass

class ConfigURL(ConfigString):
    pass

class ConfigField(object):
    def __init__(self, name, default_val, val_type,
                 field_type=ConfigType.NORMAL, category=None, flag=None,
                 nargs=None, widget_type=None,
                 widget_options=None, depends_on=None):
        self.name = name
        self.default_val = default_val
        self.val_type = val_type
        self.field_type = field_type
        self.category = category
        self.flag = flag
        self.nargs = nargs
        self.widget_type = widget_type
        if widget_options is not None:
            self.widget_options = widget_options
        else:
            self.widget_options = {}
        self.depends_on = depends_on

    def from_string(self, str_val):
        if hasattr(self.val_type, "from_string"):
            return self.val_type.from_string(str_val)
        if not str_val:
            return None
        if issubclass(self.val_type, basestring):
            return str_val
        val = ast.literal_eval(str_val)
        if not isinstance(val, self.val_type):
            raise ValueError('Output setting "%s" cannot be parsed to %s' %
                             (self.name, self.val_type.__name__))
        return val

    def to_string(self, val):
        if hasattr(self.val_type, "to_string"):
            return self.val_type.to_string(val)
        elif val is None:
            return ""
        return unicode(val)

class ConfigFieldParent(object):
    def __init__(self, name, sub_fields):
        self.name = name
        self.sub_fields = sub_fields

base_config = {
    "Command-Line":
    [ConfigField("execute", False, bool, ConfigType.COMMAND_LINE_FLAG,
                 flag='-e'),
     ConfigField("batch", False, bool, ConfigType.COMMAND_LINE_FLAG,
                 flag='-b'),
     ConfigField("outputDirectory", None, ConfigPath, flag='-o'),
     ConfigField('outputDefaultSettings', [], str,
                 ConfigType.INTERNAL_SUBOBJECT),
     ConfigField('outputSettings', [], str, ConfigType.SUBOBJECT,
                 flag='-p'),
     # ConfigField("package", [], str, flag='-p', nargs='*'),
     ConfigField("parameters", None, str, ConfigType.COMMAND_LINE),
     ConfigField("parameterExploration", False, bool,
                 ConfigType.COMMAND_LINE_FLAG),
     ConfigField('showWindow', True, bool, ConfigType.COMMAND_LINE_FLAG),
     ConfigField("withVersionTree", False, bool, ConfigType.COMMAND_LINE_FLAG),
     ConfigField("withWorkflow", False, bool, ConfigType.COMMAND_LINE_FLAG),
     ConfigField("graphsAsPdf", True, bool, ConfigType.COMMAND_LINE_FLAG)],
    "Database":
    [ConfigField("host", None, ConfigURL, ConfigType.COMMAND_LINE),
     ConfigField("port", None, int, ConfigType.COMMAND_LINE),
     ConfigField("db", None, str, ConfigType.COMMAND_LINE),
     ConfigField("user", None, str, ConfigType.COMMAND_LINE)],
    "Server":
    [ConfigField('rpcServer', None, str, ConfigType.COMMAND_LINE),
     ConfigField('rpcPort', 8080, int, ConfigType.COMMAND_LINE),
     ConfigField('rpcLogFile', os.path.join(system.vistrails_root_directory(),
                       'rpcserver.log'), ConfigPath, ConfigType.COMMAND_LINE),
     ConfigField('rpcInstances', 0, int, ConfigType.COMMAND_LINE),
     ConfigField('multithread', None, bool, ConfigType.COMMAND_LINE_FLAG),
     ConfigField('rpcConfig', os.path.join(system.vistrails_root_directory(),
                      'server.cfg'), ConfigPath, ConfigType.COMMAND_LINE)],
    "General":
    [ConfigField('autoSave', True, bool, ConfigType.ON_OFF),
     ConfigField('dbDefault', False, bool, ConfigType.ON_OFF),
     ConfigField('cache', True, bool, ConfigType.ON_OFF),
     ConfigField('stopOnError', True, bool, ConfigType.ON_OFF),
     ConfigField('executionLog', True, bool, ConfigType.ON_OFF),
     ConfigField('errorLog', True, bool, ConfigType.ON_OFF),
     ConfigField('defaultFileType', system.vistrails_default_file_type(), str,
                 widget_type="combo",
                 widget_options={"allowed_values": [".vt", ".xml"],
                                 "label": "Default File Type/Extension"}),
     ConfigField('debugLevel', 0, int,
                 flag='-v',
                 widget_type="combo",
                 widget_options={"allowed_values": [0,1,2],
                                 "label": "Show alerts for",
                                 "remap": {0: "Critical Errors Only",
                                           1: "Critical Errors and Warnings",
                                           2: "Errors, Warnings, and " \
                                           "Debug Messages"}})],
    "Startup":
    [ConfigField('maximizeWindows', False, bool, ConfigType.ON_OFF),
     ConfigField('multiHeads', False, bool, ConfigType.ON_OFF),
     ConfigField('showSplash', True, bool, ConfigType.SHOW_HIDE)],
    "Upgrades":
    [ConfigField('upgrades', True, bool, ConfigType.ON_OFF),
     ConfigField('migrateTags', False, bool, ConfigType.ON_OFF,
                 depends_on="upgrades"),
     ConfigField('hideUpgrades', True, bool, ConfigType.ON_OFF,
                 depends_on='upgrades'),
     ConfigField('upgradeDelay', True, bool, ConfigType.ON_OFF,
                 depends_on="upgrades"),
     ConfigField('upgradeModuleFailPrompt', True, bool, ConfigType.ON_OFF,
                 depends_on="upgrades")],
    "Interface":
    [ConfigField('autoConnect', True, bool, ConfigType.ON_OFF),
     ConfigField('detachHistoryView', False, bool, ConfigType.ON_OFF),
     ConfigField('showConnectionErrors', False, bool, ConfigType.SHOW_HIDE),
     ConfigField('showVariantErrors', True, bool, ConfigType.SHOW_HIDE),
     ConfigField('showDebugPopups', False, bool, ConfigType.SHOW_HIDE),
     ConfigField('showScrollbars', True, bool, ConfigType.SHOW_HIDE),
     ConfigField('showInlineParameterWidgets', False, bool, ConfigType.SHOW_HIDE),
     ConfigFieldParent('shell',
        [ConfigField('fontFace', system.shell_font_face(), str),
         ConfigField('fontSize', system.shell_font_size(), int)]),
     ConfigField('maxRecentVistrails', 5, int),
     ConfigField('viewOnLoad', "appropriate", str, widget_type='combo',
                 widget_options={"allowed_values": ["appropriate",
                                                    "history",
                                                    "pipeline"],
                                 "label": "Default view after loading vistrail:",
                                 "remap": {"appropriate": "Most Appropriate",
                                           "history": "Always History",
                                           "pipeline": "Always Pipeline"}}),
     ConfigField('enableCustomVersionColors', False, bool, ConfigType.ON_OFF),
     ConfigField('fixedCustomVersionColorSaturation', False,
                 bool, ConfigType.ON_OFF)],
    "Thumbnails":
    [ConfigFieldParent('thumbs',
        [ConfigField('autoSave', True, bool, ConfigType.ON_OFF),
         ConfigField('mouseHover', False, bool, ConfigType.ON_OFF),
         ConfigField('tagsOnly', False, bool, ConfigType.ON_OFF),
         ConfigField('cacheDir', "thumbs", ConfigPath,
                     ConfigType.NORMAL),
         ConfigField('cacheSize', 20, int, widget_type='thumbnailcache')])],
    "Packages":
    [ConfigField('enablePackagesSilently', False, bool, ConfigType.ON_OFF),
     ConfigField('loadPackages', True, bool, ConfigType.ON_OFF),
     ConfigField('installBundles', True, bool, ConfigType.ON_OFF),
     ConfigField('installBundlesWithPip', False, bool, ConfigType.ON_OFF,
                 depends_on="installBundles"),
     ConfigField('repositoryLocalPath', None, ConfigPath),
     ConfigField('repositoryHTTPURL', "http://www.vistrails.org/packages",
                 ConfigURL)],
    "Paths":
    [ConfigField('dotVistrails', system.default_dot_vistrails(),
                 ConfigPath, flag="-S"),
     ConfigField('subworkflowsDir', "subworkflows", ConfigPath),
     ConfigField('dataDir', None, ConfigPath),
     ConfigField('packageDir', None, ConfigPath),
     ConfigField('userPackageDir', "userpackages", ConfigPath),
     ConfigField('fileDir', None, ConfigPath),
     ConfigField('logDir', "logs", ConfigPath),
     ConfigField('temporaryDir', None,  ConfigPath)],
    "Advanced":
    [ConfigField('singleInstance', True, bool, ConfigType.ON_OFF),
     ConfigField('staticRegistry', None, ConfigPath)],
    "Web Sharing":
    [ConfigField('webRepositoryURL', "http://www.crowdlabs.org", ConfigURL),
     ConfigField('webRepositoryUser', None, str)],
    "Internal":
    [ConfigField('recentVistrailList', None, str, ConfigType.STORAGE),
     ConfigField('isInServerMode', False, bool, ConfigType.INTERNAL),
     ConfigField('isRunningGUI', True, bool, ConfigType.INTERNAL),
     ConfigField('spawned', False, bool, ConfigType.INTERNAL),
     ConfigField('rootDirectory', None, ConfigPath, ConfigType.INTERNAL),
     ConfigField('developerDebugger', False, bool, ConfigType.INTERNAL),
     ConfigField('dontUnloadModules', False, bool, ConfigType.INTERNAL),
     ConfigField('bundleDeclinedList', '', str, ConfigType.INTERNAL)],
    "Jobs":
    [ConfigField('jobCheckInterval', 600, int),
     ConfigField('jobAutorun', False, bool),
     ConfigField('jobList', False, bool, ConfigType.COMMAND_LINE_FLAG),
     ConfigField('jobInfo', False, bool, ConfigType.COMMAND_LINE_FLAG)],
}

# FIXME make sure that the platform-specific configs are added!
mac_config = {
    "Interface":
    [ConfigField('useMacBrushedMetalStyle', True, bool, ConfigType.ON_OFF)]
}

win_config = { }

linux_config = {
    "General":
    [ConfigField('handlerCheck', None, str, ConfigType.INTERNAL,
                 widget_type="linuxext",
                 widget_options={'label': 'Extension Handler'}),
     ConfigField('handlerDontAsk', None, bool)]
}

all_configs = [base_config, mac_config, win_config, linux_config]

def build_config_obj(d):
    new_d = {}
    for category, fields in d.iteritems():
        for field in fields:
            if isinstance(field, ConfigFieldParent):
                new_d[field.name] = build_config_obj({category:
                                                      field.sub_fields})
            elif (field.field_type == ConfigType.SUBOBJECT or
                  field.field_type == ConfigType.INTERNAL_SUBOBJECT):
                new_d[field.name] = ConfigurationObject()
            else:
                v = field.default_val
                if v is None:
                    val_type = field.val_type
                    if (field.val_type == ConfigPath or
                        field.val_type == ConfigURL):
                        val_type = basestring
                    new_d[field.name] = (v, val_type)
                else:
                    new_d[field.name] = v
    return ConfigurationObject(**new_d)

def get_system_config():
    config = {}
    config.update((k, copy.copy(v)) for k, v in base_config.iteritems())
    if system.systemType in ['Windows', 'Microsoft']:
        sys_config = win_config
    elif system.systemType in ['Linux']:
        sys_config = linux_config
    elif system.systemType in ['Darwin']:
        sys_config = mac_config
    else:
        return config
    for category, fields in sys_config.iteritems():
        if category not in base_config:
            config[category] = fields
        else:
            config[category].extend(fields)
    return config

def default():
    config = get_system_config()
    retval = build_config_obj(config)
    return retval

def parse_documentation():
    line_iter = iter(_documentation.splitlines())
    line_iter.next()
    for line in line_iter:
        arg_path, arg_type = line.strip().split(':', 1)
        doc_lines = []
        line = line_iter.next()
        while True:
            line = line_iter.next()
            if not line.strip():
                break
            doc_lines.append(line.strip())
        _docs[arg_path] = (arg_type, ' '.join(doc_lines))

def parse_simple_docs():
    line_iter = iter(_simple_documentation.splitlines())
    line = line_iter.next()
    for line in line_iter:
        (arg, doc) = line.strip().split(':', 1)
        _simple_docs[arg] = doc.strip()

def find_help(arg_path):
    if len(_docs) == 0:
        parse_documentation()

    if arg_path in _docs:
        return _docs[arg_path][1]
    return None

def find_simpledoc(arg_path):
    if len(_simple_docs) == 0:
        parse_simple_docs()

    if arg_path in _simple_docs:
        return _simple_docs[arg_path]
    return find_help(arg_path)

def set_field_labels(fields, prefix=""):
    for field in fields:
        if isinstance(field, ConfigFieldParent):
            prefix = "%s%s." % (prefix, field.name)
            set_field_labels(field.sub_fields, prefix=prefix)
        else:
            full_field_name = "%s%s" % (prefix, field.name)
            label = find_simpledoc(full_field_name)
            if label is not None and 'label' not in field.widget_options:
                field.widget_options['label'] = label

for config in all_configs:
    for field_list in config.itervalues():
        set_field_labels(field_list)

class VisTrailsHelpFormatter(argparse.HelpFormatter):
    def add_usage(self, usage, actions, groups, prefix=None):
        new_actions = []
        new_actions.append(argparse.Action([], dest="__nowhere__",
                                           metavar="[CONFIGURATION OPTIONS]"))
        for action in actions:
            if action.dest in _usage_args:
                new_actions.append(action)
        argparse.HelpFormatter.add_usage(self, usage, new_actions, groups,
                                         prefix)

class NestedSetKwargs(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        orig_dest = self.dest
        nesting = self.dest.split('.')
        while len(nesting) > 1:
            namespace = getattr(namespace, nesting.pop(0))
        self.dest = nesting[0]

        param_splitter = shlex.shlex(values, posix=True)
        param_splitter.whitespace = ','
        param_splitter.whitespace_split = True
        for param in param_splitter:
            key_val_split = param.split('=',1)
            if len(key_val_split) < 2:
                raise ValueError("Expected key=value comma-separated list")
            key, val = key_val_split

            dest = "%s.%s" % (self.dest, key)
            namespace.set_deep_value(dest, val, True)
        self.dest = orig_dest

def nested_action(parser, action_type):
    cls = parser._registry_get('action', action_type)
    if cls is None:
        raise ValueError('Action type "%s" is not defined' % action_type)

    def __call__(self, parser, namespace, values, option_string=None):
        orig_dest = self.dest
        nesting = self.dest.split('.')
        while len(nesting) > 1:
            namespace = getattr(namespace, nesting.pop(0))
        self.dest = nesting[0]
        cls.__call__(self, parser, namespace, values, option_string=None)
        self.dest = orig_dest

    nested_name = "_Nested%s" % cls.__name__[1:]
    nested_cls = type(nested_name, (cls,), {"__call__": __call__})
    return nested_cls

class RawVersionAction(argparse.Action):
    """Variant of the default _VersionAction that doesn't reflow.
    """
    def __init__(self, option_strings, version,
                 dest=argparse.SUPPRESS, default=argparse.SUPPRESS,
                 help="show program's version and exit"):
        argparse.Action.__init__(self, option_strings=option_strings,
                                 dest=dest, default=default, nargs=0,
                                 help=help)
        self.version = version

    def __call__(self, parser, namespace, values, option_string=None):
        parser.exit(message=self.version)

def build_command_line_parser(d, parser=None, prefix="", **parser_args):
    # if k is not a command-line-option, skip
    # if k is show/hide, add --show-, --hide- options
    # if k is an on/off, add --option, --no-option flags
    # otherwise, run with k converted to dashed form and store res


    def camel_to_dashes(s):
        # from http://stackoverflow.com/questions/1175208/elegant-python-function-to-convert-camelcase-to-camel-case/1176023#1176023
        s1 = re.sub('(.)([A-Z][a-z]+)', r'\1-\2', s)
        return re.sub('([a-z0-9])([A-Z])', r'\1-\2', s1).lower()


    if parser is None:
        parser = argparse.ArgumentParser(prog='vistrails',
                                         **parser_args)
        parser._my_arg_groups = {}
        parser.add_argument('vistrails', metavar='vistrail', type=str,
                            nargs='*', help="Vistrail to open")
        _usage_args.add('vistrails')
        parser.add_argument('--version', action=RawVersionAction,
                            version=system.about_string())


    prefix_dashes = ''
    if prefix:
        prefix_dashes = camel_to_dashes(prefix.replace('.', '-'))

    for category, fields in d.iteritems():
        if category == "Internal":
            # don't deal with these
            continue
        if category == "Command-Line":
            cat_group = parser
        else:
            if category not in parser._my_arg_groups:
                parser._my_arg_groups[category] = \
                                        parser.add_argument_group(category)
            cat_group = parser._my_arg_groups[category]

        for field in fields:
            if isinstance(field, ConfigFieldParent):
                build_command_line_parser({category: field.sub_fields},
                                          parser,
                                          '%s%s.' % (prefix, field.name),
                                          **parser_args)
                continue
            k_dashes = camel_to_dashes(field.name)
            help_str = find_help('%s%s' % (prefix, field.name))
            dest_name = prefix + field.name

            config_type = field.field_type
            if config_type is None:
                config_type = ConfigType.NORMAL
            if (config_type == ConfigType.INTERNAL or \
                config_type == ConfigType.STORAGE or
                config_type == ConfigType.INTERNAL_SUBOBJECT):
                # these are not in the command line
                continue
            elif config_type == ConfigType.ON_OFF:
                k_dashes = camel_to_dashes(field.name)
                group = cat_group.add_mutually_exclusive_group()
                group.add_argument('--%s%s' % (prefix_dashes, k_dashes),
                                   # action="store_true",
                                   action=nested_action(group, "store_true"),
                                   dest=dest_name, help=help_str,
                                   default=argparse.SUPPRESS)
                group.add_argument('--no-%s%s' % (prefix_dashes, k_dashes),
                                   # action="store_false",
                                   action=nested_action(group, "store_false"),
                                   dest=dest_name,
                                   help=("Inverse of --%s%s" %
                                         (prefix_dashes, k_dashes)),
                                   default=argparse.SUPPRESS)
            elif config_type == ConfigType.SHOW_HIDE:
                k_dashes = camel_to_dashes(field.name[4:])
                group = cat_group.add_mutually_exclusive_group()
                group.add_argument('--show-%s%s' % (prefix_dashes, k_dashes),
                                   # action="store_true",
                                   action = nested_action(group, "store_true"),
                                   dest=dest_name, help=help_str,
                                   default=argparse.SUPPRESS)
                group.add_argument('--hide-%s%s' % (prefix_dashes, k_dashes),
                                   # action="store_false",
                                   action=nested_action(group, "store_false"),
                                   dest=dest_name,
                                   help=("Inverse of --show-%s%s" %
                                         (prefix_dashes, k_dashes)),
                                   default=argparse.SUPPRESS)
            else:
                k_dashes = camel_to_dashes(field.name)
                long_arg = '--%s%s' % (prefix_dashes, k_dashes)
                if field.flag is not None:
                    args = (field.flag, long_arg)
                else:
                    args = (long_arg,)
                kwargs = {'dest': dest_name,
                          'help': help_str,
                          'default': argparse.SUPPRESS}
                if (field.val_type != ConfigPath and
                    field.val_type != ConfigURL and
                    field.val_type != str and
                    field.val_type != bool):
                    kwargs["type"] = field.val_type
                if config_type == ConfigType.SUBOBJECT:
                    kwargs["action"] = NestedSetKwargs
                elif config_type == ConfigType.COMMAND_LINE_FLAG:
                    kwargs["action"] = nested_action(cat_group, "store_true")
                else:
                    kwargs["action"] = nested_action(cat_group, "store")
                if field.val_type == ConfigPath:
                    kwargs["metavar"] = "DIR"
                elif field.val_type == ConfigURL:
                    kwargs["metavar"] = "URL"
                if field.nargs is not None:
                    kwargs["nargs"] = field.nargs

                cat_group.add_argument(*args, **kwargs)
                if category == "Command-Line":
                    _usage_args.add(field.name)
    return parser

def build_default_parser():
    parser_args = {"formatter_class": VisTrailsHelpFormatter,
                   "argument_default": argparse.SUPPRESS}
    return build_command_line_parser(get_system_config(), **parser_args)

def build_sphinx_parser():
    # FIXME add system-specific config options somehow
    return build_command_line_parser(base_config)

class ConfigValue(object):
    @staticmethod
    def create(value):
        if isinstance(value, bool):
            obj = ConfigBool()
        elif isinstance(value, basestring):
            obj = ConfigStr()
        elif isinstance(value, int):
            obj = ConfigInt()
        elif isinstance(value, float):
            obj = ConfigFloat()
        elif isinstance(value, list):
            obj = ConfigList()
        elif isinstance(value, ConfigurationObject):
            obj = value
        elif value is None:
            obj = None
        else:
            raise Exception('Cannot create ConfigValue from value "%s"' % value)
        if obj is not None:
            obj.set_value(value)
        return obj

    def get_value(self):
        return self.db_value

    def set_value(self, val):
        self.db_value = val

class ConfigBool(DBConfigBool, ConfigValue):
    def __copy__(self):
        """ __copy__() -> ConfigBool - Returns a clone of itself """
        return ConfigBool.do_copy(self)

    def do_copy(self, new_ids=False, id_scope=None, id_remap=None):
        cp = DBConfigBool.do_copy(self, new_ids, id_scope, id_remap)
        cp.__class__ = ConfigBool
        return cp

    @staticmethod
    def convert(_val):
        _val.__class__ = ConfigBool

    def get_value(self):
        return self.db_value.lower() == "true"

    def set_value(self, val):
        self.db_value = unicode(val)

class ConfigInt(DBConfigInt, ConfigValue):
    def __copy__(self):
        """ __copy__() -> ConfigInt - Returns a clone of itself """
        return ConfigInt.do_copy(self)

    def do_copy(self, new_ids=False, id_scope=None, id_remap=None):
        cp = DBConfigInt.do_copy(self, new_ids, id_scope, id_remap)
        cp.__class__ = ConfigInt
        return cp

    @staticmethod
    def convert(_val):
        _val.__class__ = ConfigInt

class ConfigStr(DBConfigStr, ConfigValue):
    def __copy__(self):
        """ __copy__() -> ConfigStr - Returns a clone of itself """
        return ConfigStr.do_copy(self)

    def do_copy(self, new_ids=False, id_scope=None, id_remap=None):
        cp = DBConfigStr.do_copy(self, new_ids, id_scope, id_remap)
        cp.__class__ = ConfigStr
        return cp

    @staticmethod
    def convert(_val):
        _val.__class__ = ConfigStr

class ConfigFloat(DBConfigFloat, ConfigValue):
    def __copy__(self):
        """ __copy__() -> ConfigFloat - Returns a clone of itself """
        return ConfigFloat.do_copy(self)

    def do_copy(self, new_ids=False, id_scope=None, id_remap=None):
        cp = DBConfigFloat.do_copy(self, new_ids, id_scope, id_remap)
        cp.__class__ = ConfigFloat
        return cp

    @staticmethod
    def convert(_val):
        _val.__class__ = ConfigFloat

class ConfigList(DBConfigStr, ConfigValue):
    def __copy__(self):
        """ __copy__() -> ConfigList - Returns a clone of itself """
        return ConfigList.do_copy(self)

    def do_copy(self, new_ids=False, id_scope=None, id_remap=None):
        cp = DBConfigStr.do_copy(self, new_ids, id_scope, id_remap)
        cp.__class__ = ConfigList
        return cp

    @staticmethod
    def convert(_val):
        _val.__class__ = ConfigList

    def get_value(self):
        return ast.literal_eval(self.db_value)

    def set_value(self, val):
        self.db_value = unicode(val)

class ConfigKey(DBConfigKey):
    def __init__(self, name, value):
        if isinstance(value, tuple):
            DBConfigKey.__init__(self, name=name)
            self.set_type(value[1])
        else:
            DBConfigKey.__init__(self, name=name,
                                 value=ConfigValue.create(value))
            self.set_type(type(value))

    def __copy__(self):
        """ __copy__() -> ConfigKey - Returns a clone of itself """
        return ConfigKey.do_copy(self)

    def do_copy(self, new_ids=False, id_scope=None, id_remap=None):
        cp = DBConfigKey.do_copy(self, new_ids, id_scope, id_remap)
        cp.__class__ = ConfigKey
        cp._type = self._type
        return cp

    @staticmethod
    def convert(_key):
        _key.__class__ = ConfigKey
        if isinstance(_key.db_value, DBConfiguration):
            ConfigurationObject.convert(_key.db_value)
        elif isinstance(_key.db_value, DBConfigBool):
            ConfigBool.convert(_key.db_value)
        elif isinstance(_key.db_value, DBConfigStr):
            ConfigStr.convert(_key.db_value)
        elif isinstance(_key.db_value, DBConfigInt):
            ConfigInt.convert(_key.db_value)
        elif isinstance(_key.db_value, DBConfigFloat):
            ConfigFloat.convert(_key.db_value)
        #FIXME add ConfigList to db and here
        _key.set_type(type(_key.value))

    def _get_value(self):
        if self.db_value is not None:
            return self.db_value.get_value()
        return None
    def _set_value(self, val):
        if not self.check_type(val):
            raise TypeError('Value "%s" does not match type %s' %
                            (val, self._type))
        self.db_value = ConfigValue.create(val)
    value = property(_get_value, _set_value)

    def set_type(self, t):
        if issubclass(t, basestring):
            t = basestring
        self._type = t

    def check_type(self, val):
        return val is None or isinstance(val, self._type)

class ConfigurationObject(DBConfiguration):
    """A ConfigurationObject is an InstanceObject that respects the
    following convention: values that are not 'present' in the object
    should have value (None, type), where type is the type of the
    expected object.

    ConfigurationObject exists so that the GUI can automatically infer
    the right types for the widgets.

    """


    def __init__(self, **kwargs):
        self._in_init = True
        self._unset_keys = {}
        DBConfiguration.__init__(self)
        self._in_init = False
        for k, v in kwargs.iteritems():
            if type(v) == tuple:
                self._unset_keys[k] = v
            else:
                key = ConfigKey(name=k, value=v)
                self.db_add_config_key(key)

        # InstanceObject.__init__(self, *args, **kwargs)
        self._subscribers = {}
        self.vistrails = []

    def __copy__(self):
        """ __copy__() -> ConfigurationObject - Returns a clone of itself """
        return ConfigurationObject.do_copy(self)

    def do_copy(self, new_ids=False, id_scope=None, id_remap=None):
        cp = DBConfiguration.do_copy(self, new_ids, id_scope, id_remap)
        cp._in_init = False
        cp.__class__ = ConfigurationObject
        cp._unset_keys = copy.copy(self._unset_keys)
        cp._subscribers = copy.copy(self._subscribers)
        cp.vistrails = copy.copy(self.vistrails)
        return cp

    @staticmethod
    def convert(_config_obj):
        _config_obj._in_init = False
        _config_obj.__class__ = ConfigurationObject
        for _key in _config_obj.db_config_keys:
            ConfigKey.convert(_key)
        _config_obj._subscribers = {}
        _config_obj._unset_keys = {}
        _config_obj.vistrails = []

    def get_value(self):
        return self

    def set_value(self, val):
        # it itself is the value already
        pass

    def matches_type(self, value, t):
        if t == str:
            t = basestring
        return isinstance(value, t)

    def __getattr__(self, name):
        try:
            return object.__getattribute__(self, name)
        except AttributeError:
            try:
                return self.get(name)
            except:
                raise AttributeError(name)

    def __setattr__(self, name, value):
        if name == '_subscribers' or name == '_unset_keys' or name == '_in_init' or name == 'is_dirty' or name == 'vistrails' or self._in_init:
            object.__setattr__(self, name, value)
        else:
            if name in self.db_config_keys_name_index:
                config_key = self.db_config_keys_name_index[name]
                if value is None:
                    self.db_delete_config_key(config_key)
                    self._unset_keys[name] = (None, type(config_key.value))
                else:
                    config_key.value = value
            else:
                if name not in self._unset_keys:
                    self._unset_keys[name] = (None, type(value))
                    # raise AttributeError('Key "%s" was not defined when '
                    #                      'ConfigurationObject was created' %
                    #                      name)
                if (value is not None and
                      not self.matches_type(value, self._unset_keys[name][1])):
                    raise TypeError('Value "%s" does match type "%s" for "%s"' %
                                    (value, self._unset_keys[name][1], name))
                if value is not None:
                    del self._unset_keys[name]
                    config_key = ConfigKey(name=name, value=value)
                    self.db_add_config_key(config_key)
            if name in self._subscribers:
                to_remove = []
                for subscriber in self._subscribers[name]:
                    obj = subscriber()
                    if obj:
                        obj(name, value)
                    else:
                        to_remove.append(obj)
                for ref in to_remove:
                    self._subscribers[name].remove(ref)

    def __eq__(self, other):
        if type(self) != type(other):
            return False
        seen_keys = set()
        for name in self.keys():
            if self.is_unset(name):
                continue
            seen_keys.add(name)
            if name not in other.keys():
                return False
            val1 = getattr(self, name)
            val2 = getattr(other, name)
            if type(val1) != type(val2):
                return False
            if val1 != val2:
                return False
        for name in other.keys():
            if other.is_unset(name):
                continue
            if name not in seen_keys:
                return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def __contains__(self, k):
        return self.has(k)

    def __getitem__(self, k):
        return self.get(k)

    def __setitem__(self, k, v):
        return self.__setattr__(k, v)

    def update(self, other):
        for name, other_key in other.db_config_keys_name_index.iteritems():
            if (isinstance(other_key.value, ConfigurationObject) and
                    self.has(name)):
                self.get(name).update(other_key.value)
            else:
                self.__setattr__(name, other_key.value)

    def unsubscribe(self, field, callable_):
        """unsubscribe(field, callable_): remove observer from subject
        """
        self._subscribers[field].remove(weakref.ref(callable_))

    def subscribe(self, field, callable_):
        """subscribe(field, callable_): call observer callable_ when
        self.field is set.
        """
        append_to_dict_of_lists(self._subscribers, field,
                                Ref(callable_))

    def has(self, key):
        """has(key) -> bool.

        Returns true if key has valid value in the object.
        """

        return key in self.db_config_keys_name_index

    def get(self, key):
        if key in self._unset_keys:
            return self._unset_keys[key]
        config_key = self.db_config_keys_name_index[key]
        return config_key.value

    def is_unset(self, keys_str):
        keys = keys_str.split('.')
        config = self
        for key in keys[:-1]:
            config = config.get(key)
        return keys[-1] in config._unset_keys

    def get_deep_value(self, keys_str):
        # keys_str is something like "thmbs.cacheDir"
        keys = keys_str.split('.')
        config = self
        for key in keys:
            config = config.get(key)
        return config

    def has_deep_value(self, keys_str):
        # keys_str is something like "thmbs.cacheDir"
        keys = keys_str.split('.')
        config = self
        for key in keys:
            if config.has(key):
                config = config.get(key)
            else:
                return False
        return True

    def set_deep_value(self, keys_str, value, create_if_missing=False):
        keys = keys_str.split('.')
        config = self
        for key in keys[:-1]:
            if config.has(key):
                config = config.get(key)
            elif create_if_missing:
                new_config = ConfigurationObject()
                config.__setattr__(key, new_config)
                config = new_config
            else:
                raise ValueError('No path for key "%s"' % keys_str)
        config.__setattr__(keys[-1], value)

    def check(self, key):
        """check(key) -> obj

        Returns False if key is absent in object, and returns object
        otherwise.
        """

        return self.has(key) and getattr(self, key)

    def allkeys(self):
        """allkeys() -> list of strings

        Returns all options stored in this object.
        """

        return self.db_config_keys_name_index.keys() + self._unset_keys.keys()

    def keys(self):
        """keys(self) -> list of strings
        Returns all public options stored in this object.
        Public options are keys that do not start with a _
        """
        return [k for k in itertools.chain(self.db_config_keys_name_index,
                                           self._unset_keys)
                if not k.startswith('_')]

# def default():
#     """ default() -> ConfigurationObject
#     Returns the default configuration of VisTrails

#     """

#     base_dir = {
#         'abstractionsDirectory': os.path.join("$DOT_VISTRAILS",
#                                               "subworkflows"),
#         'alwaysShowDebugPopup': False,
#         'autoConnect': True,
#         'autosave': True,
#         'dataDirectory': (None, str),
#         'dbDefault': False,
# #        'debugSignals': False,
#         'defaultFileType':system.vistrails_default_file_type(),
#         'detachHistoryView': False,
#         'dotVistrails': system.default_dot_vistrails(),
#         'enablePackagesSilently': False,
#         'errorOnConnectionTypeerror': False,
#         'errorOnVariantTypeerror': True,
#         'executeWorkflows': False,
#         'fileDirectory': (None, str),
#         'fixedSpreadsheetCells': False,
# #        'evolutionGraph': (None, str),
#         'installBundles': True,
#         'installBundlesWithPip': False,
#         'interactiveMode': True,
#         'logFile': os.path.join("$DOT_VISTRAILS", "vistrails.log"),
#         'logger': default_logger(),
#         'maxMemory': (None, int),
#         'maximizeWindows': False,
#         'maxRecentVistrails': 5,
#         'migrateTags': False,
#         'minMemory': (None, int),
#         'multiHeads': False,
#         'nologger': False,
#         'packageDirectory': (None, str),
#         'pythonPrompt': False,
#         'recentVistrailList': (None, str),
#         'repositoryLocalPath': (None, str),
#         'repositoryHTTPURL': "http://www.vistrails.org/packages",
#         'reviewMode': False,
#         'rootDirectory': (None, str),
#         'runningJobsList': (None, str),
#         'shell': default_shell(),
#         'showScrollbars': True,
#         'showMovies': True,
#         'showSplash': True,
#         'showSpreadsheetOnly': False,
#         'singleInstance': True,
#         'spreadsheetDumpCells': (None, str),
#         'spreadsheetDumpPDF': False,
#         'staticRegistry': (None, str),
#         'stopOnError': True,
#         'temporaryDirectory': (None, str),
#         'thumbs': default_thumbs(),
#         'upgradeOn': True,
#         'upgradeDelay': True,
#         'upgradeModuleFailPrompt': True,
#         'useCache': True,
#         'userPackageDirectory': os.path.join("$DOT_VISTRAILS", "userpackages"),
#         'verbosenessLevel': (None, int),
# #        'workflowGraph': (None, str),
# #        'workflowInfo': (None, str),
#         'webRepositoryLogin': (None, str),
#         'webRepositoryURL': "http://www.crowdlabs.org",
#         'isInServerMode': False,
#         }
#     specific_dir = add_specific_config(base_dir)
#     return ConfigurationObject(**specific_dir)

# def default_logger():
#     """default_logger() -> ConfigurationObject
#     Returns the default configuration for the VisTrails logger

#     """
#     logger_dir = {
#         'dbHost': '',
#         'dbName': '',
#         'dbPasswd': '',
#         'dbPort': 0,
#         'dbUser': '',
#         }
#     return ConfigurationObject(**logger_dir)

# def default_shell():
#     """default_shell() -> ConfigurationObject
#     Returns the default configuration for the VisTrails shell
#
#     """
#     if system.systemType == 'Linux':
#         shell_dir = {
#             'font_face': 'Fixed',
#             'font_size': 12,
#             }
#     elif system.systemType in ['Windows', 'Microsoft']:
#         shell_dir = {
#             'font_face': 'Courier New',
#             'font_size': 8,
#             }
#     elif system.systemType == 'Darwin':
#         shell_dir = {
#             'font_face': 'Monaco',
#             'font_size': 12,
#             }
#     else:
#         raise VistrailsInternalError('system type not recognized')
#     return ConfigurationObject(**shell_dir)

# def default_thumbs():
#     """default_thumbs() -> ConfigurationObject
#     Returns the default configuration for VisTrails Pipelines Thumbnails
#     """
#     thumbs_dir = {
#                   'autoSave': True,
#                   'cacheDirectory': os.path.join("$DOT_VISTRAILS", "thumbs"),
#                   'cacheSize': 20,
#                   'mouseHover': False,
#                   'tagsOnly': False,
#                 }
#     return ConfigurationObject(**thumbs_dir)

def add_specific_config(base_dir):
    """add_specific_config() -> dict

    Returns a dict with other specific configuration to the current
    platform added to base_dir

    """
    newdir = dict(base_dir)
    if system.systemType == 'Darwin':
        newdir['useMacBrushedMetalStyle'] = True

    return newdir

def get_vistrails_persistent_configuration():
    """get_vistrails_persistent_configuration() -> ConfigurationObject or None
    Returns the persistent configuration of the application. It returns None if
    configuration was not found (when running as a bogus application
    for example.
    Notice that this function should be use only to write configurations to
    the user's startup.xml file. Otherwise, use get_vistrails_configuration  or
    get_vistrails_temp_configuration.

    """
    from vistrails.core.application import get_vistrails_application
    app = get_vistrails_application()
    if hasattr(app, 'configuration'):
        return app.configuration
    else:
        return None

def get_vistrails_temp_configuration():
    """get_vistrails_temp_configuration() -> ConfigurationObject or None
    Returns the temp configuration of the application. It returns None if
    configuration was not found (when running as a bogus application
    for example. The temp configuration is the one that is used just for the
    current session and is not persistent. To make changes persistent,
    use get_vistrails_persistent_configuration() instead.

    """
    from vistrails.core.application import get_vistrails_application
    app = get_vistrails_application()
    if hasattr(app, 'temp_configuration'):
        return app.temp_configuration
    else:
        return None

get_vistrails_configuration = get_vistrails_temp_configuration

import tempfile
import unittest

class TestConfiguration(unittest.TestCase):
    def test_config(self):
        conf = ConfigurationObject(a="blah", b=3.45, c=1, d=True)
        self.assertEqual(conf.a, "blah")
        self.assertAlmostEqual(conf.b, 3.45)
        self.assertEqual(conf.c, 1)
        self.assertEqual(conf.d, True)

    def test_default(self):
        conf = default()
        self.assertEqual(conf.showWindow, True)
        self.assertEqual(conf.maxRecentVistrails, 5)

    def test_has(self):
        conf = default()
        self.assertTrue(conf.has("showWindow"))
        self.assertFalse(conf.has("reallyDoesNotExist"))

    def test_check(self):
        conf = default()
        self.assertTrue(conf.check("showWindow"))
        self.assertFalse(conf.check("showDebugPopups"))
        self.assertFalse(conf.check("thumbs.mouseHover"))

    def test_update(self):
        conf1 = default()
        conf2 = ConfigurationObject(showDebugPopups=True,
                                    logDir="/tmp",
                                    thumbs=ConfigurationObject(
                                        autoSave=True,
                                        cacheDir="/tmp",
                                        cacheSize=10,
                                        mouseHover=True,
                                        tagsOnly=False))

        conf1.update(conf2)
        self.assertTrue(conf1.showDebugPopups)
        self.assertEqual(conf1.logDir, "/tmp")
        self.assertEqual(conf1.thumbs.mouseHover, True)

        conf2.showWindow = False
        self.assertTrue(conf1.showWindow)

    def test_unset_params(self):
        conf = ConfigurationObject(test_field=(None, str))
        self.assertTrue(conf.is_unset("test_field"))
        self.assertIn("test_field", conf.keys())

    def test_type_mismatch(self):
        conf = default()
        with self.assertRaises(TypeError):
            conf.showWindow = 1

        # allowing this now
        # with self.assertRaises(AttributeError):
        #     conf.reallyDoesNotExist = True

    def test_serialize(self):
        from vistrails.db.persistence import DAOList
        conf1 = default()
        (fd, fname) = tempfile.mkstemp()
        os.close(fd)
        try:
            dao = DAOList()
            dao.save_to_xml(conf1, fname, {})
            conf2 = dao.open_from_xml(fname, ConfigurationObject.vtType)
            ConfigurationObject.convert(conf2)
        finally:
            os.unlink(fname)

        self.assertEqual(conf1, conf2)

    def test_copy(self):
        conf1 = default()
        conf2 = copy.copy(conf1)
        self.assertEqual(conf1, conf2)
        self.assertItemsEqual(conf1._unset_keys.keys(),
                              conf2._unset_keys.keys())

    def test_parser(self):
        if sys.version_info < (2, 7):
            self.skipTest("argparse on Python 2.6: bug 10680")
        from vistrails.tests.utils import capture_stdout, capture_stderr
        p = build_command_line_parser(base_config)
        err = []
        try:
            with capture_stdout() as err:
                with self.assertRaises(SystemExit) as e:
                    p.parse_args(["-h"])
            self.assertEqual(e.exception.code, 0)
            self.assertGreater(len(err), 20)
        except:
            sys.stdout.write('\n'.join(err))
            raise
        try:
            with capture_stderr() as err:
                with self.assertRaises(SystemExit) as e:
                    p.parse_args(["--db-default", "--no-db-default"])
            self.assertEqual(e.exception.code, 2)
        except:
            sys.stderr.write('\n'.join(err))
            raise

    def test_parse_into_config(self):
        p = build_command_line_parser(base_config)
        config = default()
        self.assertFalse(config.dbDefault)
        p.parse_args(args=["--db-default", "--dot-vistrails", "/tmp"],
                     namespace=config)
        self.assertTrue(config.dbDefault)
        self.assertEqual(config.dotVistrails, "/tmp")

    def test_parse_output_settings(self):
        p = build_command_line_parser(base_config)
        config = default()
        p.parse_args(args=["-p", "file.series=false"],
                     namespace=config)
        self.assertTrue(config.outputSettings.has("file"))
        self.assertTrue(config.outputSettings.file.has("series"))
        self.assertEqual(config.outputSettings.file.series, "false")

    def test_multiple_params(self):
        p = build_command_line_parser(base_config)
        config = default()
        p.parse_args(args=["-p", "file.series=false",
                           "-p", "file.suffix=.png"],
                     namespace=config)
        self.assertTrue(config.outputSettings.has("file"))
        self.assertTrue(config.outputSettings.file.has("series"))
        self.assertTrue(config.outputSettings.file.has("suffix"))
        self.assertEqual(config.outputSettings.file.series, "false")
        self.assertEqual(config.outputSettings.file.suffix, ".png")

    def test_comma_sep_params(self):
        p = build_command_line_parser(base_config)
        config = default()
        p.parse_args(args=["-p", 'file.series=false,other.separator=","'],
                     namespace=config)
        self.assertTrue(config.outputSettings.has("file"))
        self.assertTrue(config.outputSettings.file.has("series"))
        self.assertEqual(config.outputSettings.file.series, "false")
        self.assertTrue(config.outputSettings.has("other"))
        self.assertTrue(config.outputSettings.other.has("separator"))
        self.assertEqual(config.outputSettings.other.separator, ",")


if __name__ == '__main__':
    unittest.main()




###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################

"""Widgets to display/edit configuration objects."""
from __future__ import division

import os
import os.path
from PyQt4 import QtGui, QtCore

from vistrails.core.configuration import ConfigurationObject, \
    ConfigFieldParent, ConfigPath, \
    get_vistrails_configuration, find_simpledoc

from vistrails.core.thumbnails import ThumbnailCache
from vistrails.gui.common_widgets import QSearchTreeWindow, QSearchTreeWidget, \
    QDirectoryChooserToolButton
from vistrails.gui.utils import YES_BUTTON, NO_BUTTON, show_question, show_warning

from vistrails.core import system

##############################################################################

def bool_conv(st):
    if st == 'True':
        return True
    elif st == 'False':
        return False
    else:
        raise TypeError('Bogus value for bool_conv ' + str(st))

class QConfigurationTreeWidgetItem(QtGui.QTreeWidgetItem):

    def __init__(self, parent, obj, parent_obj, name, temp_obj, temp_parent_obj):
        lst = [name]
        t = type(obj)
        if t == bool:
            self._obj_type = bool_conv
        else:
            self._obj_type = t
        self._parent_obj = parent_obj
        self._temp_parent_obj = temp_parent_obj
        
        self._name = name
        if t == ConfigurationObject:
            lst.extend(['', ''])
            QtGui.QTreeWidgetItem.__init__(self, parent, lst)
            self.setFlags(self.flags() & ~(QtCore.Qt.ItemIsDragEnabled |
                                           QtCore.Qt.ItemIsSelectable ))
        elif t == tuple and obj[0] is None and isinstance(obj[1], type):
            self._obj_type = obj[1]
            lst.extend(['', obj[1].__name__])
            QtGui.QTreeWidgetItem.__init__(self, parent, lst)
            self.setFlags((self.flags() & ~QtCore.Qt.ItemIsDragEnabled) |
                          QtCore.Qt.ItemIsEditable)
        else:
            lst.extend([str(obj), type(obj).__name__])
            QtGui.QTreeWidgetItem.__init__(self, parent, lst)
            self.setFlags((self.flags() & ~QtCore.Qt.ItemIsDragEnabled) |
                          QtCore.Qt.ItemIsEditable)

    def change_value(self, new_value):
        # if this is a parent ConfigurationObject, do nothing
        if self._parent_obj and not self._obj_type == ConfigurationObject:
            setattr(self._parent_obj, self._name, self._obj_type(new_value))
            setattr(self._temp_parent_obj, self._name, self._obj_type(new_value))

    def _get_name(self):
        return self._name
    name = property(_get_name)

class QConfigurationTreeWidgetItemDelegate(QtGui.QItemDelegate):
    """
    QConfigurationTreeWidgetItemDelegate allows a custom editor for
    each column of the QConfigurationTreeWidget    
    """
    
    def createEditor(self, parent, option, index):
        """ createEditor(parent: QWidget,
                         option: QStyleOptionViewItem,
                         index: QModelIndex) -> QWidget
        Return the editing widget depending on columns
        
        """
        # We only allow users to edit the  second column
        if index.column()==1:
            dataType = str(index.sibling(index.row(), 2).data())
            
            # Create the editor based on dataType
            if dataType=='int':
                editor = QtGui.QLineEdit(parent)
                editor.setValidator(QtGui.QIntValidator(parent))
            elif dataType=='bool':
                editor = QtGui.QComboBox(parent)
                editor.addItem('True')
                editor.addItem('False')
            else:
                editor = QtGui.QItemDelegate.createEditor(self, parent,
                                                          option, index)
            return editor            
        return None

    def setEditorData(self, editor, index):
        """ setEditorData(editor: QWidget, index: QModelIndex) -> None
        Set the editor to reflects data at index
        
        """
        if isinstance(editor, QtGui.QComboBox):
            editor.setCurrentIndex(editor.findText(index.data()))
        else:
            QtGui.QItemDelegate.setEditorData(self, editor, index)

    def setModelData(self, editor, model, index):
        """ setModelData(editor: QStringEdit,
                         model: QAbstractItemModel,
                         index: QModelIndex) -> None
        Set the text of the editor back to the item model
        
        """
        if isinstance(editor, QtGui.QComboBox):
            model.setData(index, editor.currentText())
        elif isinstance(editor, QtGui.QLineEdit):
            model.setData(index, editor.text())
        else:
            # Should never get here
            assert False
    

class QConfigurationTreeWidget(QSearchTreeWidget):

    def __init__(self, parent, persistent_config, temp_config):
        QSearchTreeWidget.__init__(self, parent)
        self.setMatchedFlags(QtCore.Qt.ItemIsEditable)
        self.setColumnCount(3)
        lst = ['Name', 'Value', 'Type']
        self.setHeaderLabels(lst)
        self.create_tree(persistent_config, temp_config)

    def create_tree(self, persistent_config, temp_config):
        def create_item(parent, obj, parent_obj, name, temp_obj, temp_parent_obj):
            item = QConfigurationTreeWidgetItem(parent, obj, parent_obj, 
                                                name, temp_obj, temp_parent_obj)
            if isinstance(obj, ConfigurationObject):
                for key in sorted(obj.keys()):
                    create_item(item, getattr(obj, key), obj, key, 
                                getattr(temp_obj, key), temp_obj)

        # disconnect() and clear() are here because create_tree might
        # also be called when an entirely new configuration object is set.

        self.disconnect(self, QtCore.SIGNAL('itemChanged(QTreeWidgetItem *, int)'),
                        self.change_configuration)
        self.clear()
        self._configuration = persistent_config
        self._temp_configuration = temp_config
        create_item(self, self._configuration, None, 'configuration', 
                    self._temp_configuration, None)

        self.expandAll()
        self.resizeColumnToContents(0)
        self.connect(self,
                     QtCore.SIGNAL('itemChanged(QTreeWidgetItem *, int)'),
                     self.change_configuration)

    def change_configuration(self, item, col):
        if item.flags() & QtCore.Qt.ItemIsEditable:
            new_value = self.indexFromItem(item, col).data()
            item.change_value(new_value)
            # option-specific code
            if item._name == 'dbDefault':
                # Update the state of the icons if changing between db and
                # file support
                dbState = getattr(get_vistrails_configuration(), 'dbDefault')
                if new_value != dbState:
                    from vistrails.gui.vistrails_window import _app
                    _app.setDBDefault(dbState)

            self.emit(QtCore.SIGNAL('configuration_changed'),
                      item, new_value)
        
class QConfigurationTreeWindow(QSearchTreeWindow):

    def __init__(self, parent, persistent_config, temp_config):
        self._configuration_object = persistent_config
        self._temp_configuration = temp_config
        QSearchTreeWindow.__init__(self, parent)

    def createTreeWidget(self):
        self.setWindowTitle('Configuration')
        treeWidget = QConfigurationTreeWidget(self, self._configuration_object,
                                              self._temp_configuration)
        
        # The delegate has to be around (self._delegate) to
        # work, else the instance will be clean by Python...
        self._delegate = QConfigurationTreeWidgetItemDelegate()
        treeWidget.setItemDelegate(self._delegate)
        return treeWidget


class QConfigurationWidget(QtGui.QWidget):

    def __init__(self, parent, persistent_config, temp_config):
        QtGui.QWidget.__init__(self, parent)
        layout = QtGui.QVBoxLayout(self)
        self.setLayout(layout)

        self._tree = QConfigurationTreeWindow(self, persistent_config,
                                              temp_config)
        lbl = QtGui.QLabel("Set configuration variables for VisTrails here.", self)
        layout.addWidget(lbl)
        layout.addWidget(self._tree)

    def configuration_changed(self, persistent_config, temp_config):
        self._tree.treeWidget.create_tree(persistent_config, temp_config)

class QConfigurationWidgetItem(object):
    def __init__(self, key, field, callback_f):
        self.key = key
        self.field = field
        self.change_callback_f = callback_f
        self._desc = None

    def get_desc(self):
        if self._desc is not None:
            return self._desc

        options = self.get_widget_options()
        if "label" in options:
            return options["label"]
        return ""

    def set_desc(self, desc=None):
        self._desc = desc

    def get_label_text(self):
        return self.get_desc()

    def set_value(self, value, signal=True):
        raise NotImplementedError("Subclass needs to implement this method")

    def value_changed(self, value):
        self.change_callback_f(self, self.key, self.field, value)

    def get_widget_options(self):
        options = {}
        if self.field.widget_options is not None:
            options = self.field.widget_options
        return options

class QConfigurationCheckBox(QtGui.QCheckBox, QConfigurationWidgetItem):
    def __init__(self, key, field, callback_f, parent=None):
        QtGui.QCheckBox.__init__(self, parent)
        QConfigurationWidgetItem.__init__(self, key, field, callback_f)
        self.setText(self.get_desc())
        self.toggled.connect(self.value_changed)

    def set_value(self, value, signal=True):
        if not signal:
            self.toggled.disconnect(self.value_changed)
        self.setChecked(value)
        if not signal:
            self.toggled.connect(self.value_changed)

    def get_label_text(self):
        return ""

class QConfigurationLineEdit(QtGui.QLineEdit, QConfigurationWidgetItem):
    def __init__(self, key, field, callback_f, parent=None):
        QtGui.QLineEdit.__init__(self, parent)
        QConfigurationWidgetItem.__init__(self, key, field, callback_f)
        self.setMinimumWidth(200)
        self.editingFinished.connect(self.value_changed)

    def value_changed(self):
        QConfigurationWidgetItem.value_changed(self, self.text())

    def set_value(self, value, signal=True):
        if value is None:
            value = ""
        if not signal:
            self.editingFinished.disconnect(self.value_changed)
        self.setText(unicode(value))
        if not signal:
            self.editingFinished.connect(self.value_changed)

class QConfigurationLineEditButton(QtGui.QWidget, QConfigurationWidgetItem):
    def __init__(self, key, field, callback_f, button, parent=None):
        QtGui.QWidget.__init__(self, parent)
        QConfigurationWidgetItem.__init__(self, key, field, callback_f)

        layout = QtGui.QHBoxLayout()
        layout.setMargin(0)
        layout.setSpacing(5)

        self.line_edit = QtGui.QLineEdit()
        self.line_edit.setMinimumWidth(200)
        layout.addWidget(self.line_edit)

        if button is not None:
            layout.addWidget(button)
        self.setLayout(layout)

        self.line_edit.editingFinished.connect(self.value_changed)

    def add_button(self, button):
        self.layout().addWidget(button)

    def value_changed(self):
        QConfigurationWidgetItem.value_changed(self, self.line_edit.text())

    def set_value(self, value, signal=True):
        if value is None:
            value = ""
        if not signal:
            self.line_edit.editingFinished.disconnect(self.value_changed)
        self.line_edit.setText(unicode(value))
        if not signal:
            self.line_edit.editingFinished.connect(self.value_changed)

class QConfigurationPathEdit(QConfigurationLineEditButton):
    def __init__(self, key, field, callback_f, 
                 button_cls=QDirectoryChooserToolButton, parent=None):
        QConfigurationLineEditButton.__init__(self, key, field, callback_f,
                                              None, parent)
        button = button_cls(self, self.line_edit)
        self.add_button(button)

class QConfigurationThumbnailCache(QConfigurationLineEditButton):
    def __init__(self, key, field, callback_f, parent=None):
        button = QtGui.QPushButton("Clear...")
        button.setAutoDefault(False)
        button.clicked.connect(self.clear_clicked)
        QConfigurationLineEditButton.__init__(self, key, field, callback_f, 
                                              button, parent)

    def clear_clicked(self, checked=False):
        thumbnail_dir = system.get_vistrails_directory("thumbs.cacheDir")
        res = show_question('VisTrails',
                            ("All files in %s will be removed. "
                             "Are you sure? " % thumbnail_dir),
                            buttons = [YES_BUTTON,NO_BUTTON],
                            default = NO_BUTTON)
        if res == YES_BUTTON:
            ThumbnailCache.getInstance().clear()

class QConfigurationLabelButton(QtGui.QWidget, QConfigurationWidgetItem):
    def __init__(self, key, field, callback_f, label=None, button=None, 
                 parent=None):
        QtGui.QWidget.__init__(self, parent)
        QConfigurationWidgetItem.__init__(self, key, field, callback_f)

        layout = QtGui.QHBoxLayout()
        layout.setMargin(0)
        layout.setSpacing(5)

        if label is not None:
            self.label = label
            layout.addWidget(self.label)

        if button is not None:
            self.button = button
            layout.addWidget(self.button)
        self.setLayout(layout)

    def add_button(self, button):
        self.button = button
        self.layout().addWidget(self.button)

    def add_label(self, label):
        self.label = label
        self.layout().insertWidget(0, self.label)

    def set_value(self, value, signal=True):
        # nothing to do here
        pass

class QConfigurationLinuxHandler(QConfigurationLabelButton):
    def __init__(self, key, field, callback_f, parent=None):
        from vistrails.gui.application import linux_default_application_set
        if linux_default_application_set():
            label = QtGui.QLabel(".vt, .vtl handlers installed")
            button = None
        else:
            label = QtGui.QLabel(".vt, .vtl handlers not installed")
        button = QtGui.QPushButton("Install...")
        button.setAutoDefault(False)
        button.clicked.connect(self.install_clicked)
        QConfigurationLabelButton.__init__(self, key, field, callback_f, 
                                           label, button, parent)

    def install_clicked(self, checked=False):
        from vistrails.core.application import get_vistrails_application
        app = get_vistrails_application()
        if app.ask_update_default_application(False):
            self.label.setText(".vt, .vtl handlers installed")

class QConfigurationComboBox(QtGui.QComboBox, QConfigurationWidgetItem):
    def __init__(self, key, field, callback_f, parent=None):
        QtGui.QComboBox.__init__(self, parent)
        QConfigurationWidgetItem.__init__(self, key, field, callback_f)

        inv_remap = None
        options = self.get_widget_options()
        if "allowed_values" in options:
            values = options["allowed_values"]
            if "remap" in options:
                remap = options["remap"]
                inv_remap = dict((v, k) for (k, v) in remap.iteritems())
                entries = [remap[v] for v in values]
            else:
                entries = values
            for entry in entries:
                self.addItem(entry)

        self.currentIndexChanged[int].connect(self.value_changed)

    def set_value(self, value, signal=True):
        options = self.get_widget_options()
        if not signal:
            self.currentIndexChanged[int].disconnect(self.value_changed)
        if value is not None and "allowed_values" in options:
            if "remap" in options:
                remap = options["remap"]
                cur_text = remap[value]
            else:
                cur_text = value
            self.setCurrentIndex(self.findText(cur_text))
        else:
            self.setCurrentIndex(-1)
        if not signal:
            self.currentIndexChanged[int].connect(self.value_changed)

class QConfigurationPane(QtGui.QWidget):
    def __init__(self, parent, persistent_config, temp_config, cat_fields):
        QtGui.QWidget.__init__(self, parent)
        layout = QtGui.QFormLayout()
        layout.setMargin(10)
        layout.setSpacing(4)
        self.setLayout(layout)
        self._configuration = persistent_config
        self._temp_configuration = temp_config

        self._fields = {}
        self._field_layouts = {}

        for category, fields in cat_fields:
            self.process_fields(layout, fields, category)
            spacer_widget = QtGui.QWidget()
            spacer_layout = QtGui.QVBoxLayout()
            spacer_layout.setMargin(0)
            spacer_layout.addSpacing(15)
            spacer_widget.setLayout(spacer_layout)
            layout.addRow("", spacer_widget)

    def process_fields(self, layout, fields, category, parent_fields=[], 
                       prev_field=None, prefix=""):
        for field in fields:
            if isinstance(field, ConfigFieldParent):
                self.process_fields(layout, field.sub_fields, category,
                                    parent_fields, prev_field,
                                    prefix="%s%s." % (prefix, field.name))
            else:
                if field.depends_on is not None:
                    if field.depends_on not in parent_fields:
                        if field.depends_on == prev_field:
                            parent_fields.append(prev_field)
                        else:
                            raise Exception("Dependent field %s should "
                                            "follow parent." % field.name)
                    parent_idx = parent_fields.index(field.depends_on)
                    parent_fields = parent_fields[:parent_idx+1]
                    indent = 4 * len(parent_fields)
                else:
                    parent_fields = []
                    indent = 0
                self.add_field(layout, field, category, prefix=prefix, 
                               indent=indent)
                prev_field = field.name
                category = ""

    def add_field(self, base_layout, field, category="", startup_only=False,
                  prefix="", indent=0):
        label_widget = QtGui.QWidget()
        label_layout = QtGui.QHBoxLayout()
        label_layout.setMargin(0)
        label_layout.setSpacing(5)
        label_widget.setLayout(label_layout)

        config_key = "%s%s" % (prefix, field.name)
        if self._temp_configuration.is_unset(config_key):
            config_val = None
        else:
            config_val = self._temp_configuration.get_deep_value(config_key)
        if self._configuration.is_unset(config_key):
            perm_config_val = None
        else:
            perm_config_val = self._configuration.get_deep_value(config_key)

        icon = self.style().standardIcon(QtGui.QStyle.SP_MessageBoxWarning)
        label = QtGui.QLabel()
        label.setPixmap(icon.pixmap(14,14))
        label.setToolTip("This option has been changed for this session")
        label_layout.addWidget(label, 0, QtCore.Qt.AlignCenter)
            
        space = 0
        if not startup_only and config_val == perm_config_val:
            space = (label.sizeHint().width() +
                     label_layout.spacing() * (indent + 1))
            label.hide()
        elif indent > 0:
            space = label_layout.spacing() * indent

        if space > 0:
            spacer = QtGui.QSpacerItem(space, label.sizeHint().height())
            label_layout.insertSpacerItem(0, spacer)

        config_desc = find_simpledoc(config_key)
        widget_type = field.widget_type
        if widget_type is None:
            if field.val_type == bool:
                widget_type = "checkbox"
            elif field.val_type == ConfigPath:
                widget_type = "pathedit"
            else:
                widget_type = "lineedit"

        if widget_type == "combo":
            widget = QConfigurationComboBox(config_key, field,
                                            self.field_changed)
        elif widget_type == "lineedit":
            widget = QConfigurationLineEdit(config_key, field,
                                            self.field_changed)
        elif widget_type == "pathedit":
            widget = QConfigurationPathEdit(config_key, field,
                                            self.field_changed)
        elif widget_type == "thumbnailcache":
            widget = QConfigurationThumbnailCache(config_key, field,
                                                  self.field_changed)
        elif widget_type == "linuxext":
            widget = QConfigurationLinuxHandler(config_key, field,
                                                self.field_changed)
        else:
            config_val = bool(config_val)
            widget = QConfigurationCheckBox(config_key, field,
                                            self.field_changed)
        widget.set_value(config_val, False)

        label_text = widget.get_label_text()
        if not label_text and category:
            label_text = category
        if label_text:
            label = QtGui.QLabel(label_text + ":")
            label_layout.addWidget(label)

        base_layout.addRow(label_widget, widget)
        self._field_layouts[config_key] = (base_layout, base_layout.rowCount())

    def field_changed(self, widget, config_key, field, val):
        config_val = self._configuration.get_deep_value(config_key)
        if config_val != self._temp_configuration.get_deep_value(config_key):
            retval = QtGui.QMessageBox.question(
                self, 
                "Change Setting",
                "This configuration value has been temporarily changed. "
                "If you change it, it will be changed permanently.  Do you "
                "want to continue?", 
                QtGui.QMessageBox.Cancel | QtGui.QMessageBox.Ok,
                QtGui.QMessageBox.Ok)
            if retval != QtGui.QMessageBox.Ok:
                # revert widget's value
                widget.set_value(self._temp_configuration.get_deep_value(
                    config_key))
                return
            # need to update hbox to reflect change...
            form_layout, row = self._field_layouts[config_key]
            label_layout = form_layout.itemAt(row, QtGui.QFormLayout.LabelRole).widget().layout()
            leading_item = label_layout.itemAt(0)
            if isinstance(leading_item.widget(), QtGui.QLabel):
                label = leading_item.widget()
                spacer = QtGui.QSpacerItem(label.sizeHint().width() + \
                                           label_layout.spacing(),
                                           label.sizeHint().height())
                label_layout.insertSpacerItem(0, spacer)
            else:
                spacer = leading_item
                label = label_layout.itemAt(1).widget()
                spacer.changeSize((spacer.sizeHint().width() +
                                   label.sizeHint().width() +
                                   label_layout.spacing()),
                                  label.sizeHint().height())
            label.hide()
        # FIXME
        if False:
            QtGui.QMessageBox.information(
                self, "Change Setting",
                "You must restart VisTrails for this setting to take effect.")

        setattr(self._temp_configuration, config_key, val)
        setattr(self._configuration, config_key, val)

# TODO: Make sure this functionality (Move and Clear Cache) is preserved

class QThumbnailConfiguration(QtGui.QWidget):
    def thumbs_cache_directory_changed(self):
        """ thumbs_cache_changed(v: int) -> None
        
        """
        value = str(self._thumbs_cache_directory_edt.text())
        old_folder = self._configuration.thumbs.cacheDirectory
        if os.path.exists(value):
            self._configuration.thumbs.cacheDirectory = value
            self._temp_configuration.thumbs.cacheDirectory = value
            self.emit(QtCore.SIGNAL('configuration_changed'),
                      None, value)
            self._cache.move_cache_directory(old_folder,value)
        else:
            show_warning('VisTrails', 'The directory specified does not exist.')
            self._thumbs_cache_directory_edt.setText(old_folder)




###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2013-2014, NYU-Poly.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################

from __future__ import division

from datetime import datetime
from file_archive import hash_file, hash_directory
import os

import vistrails.core.debug as debug
from vistrails.core.modules.basic_modules import Directory, File, Path, \
    PathObject
from vistrails.core.modules.config import IPort, OPort, ModuleSettings
from vistrails.core.modules.vistrails_module import Module, ModuleError

from .common import KEY_TYPE, TYPE_INPUT, KEY_TIME, \
    get_default_store, PersistentHash
from .queries import Metadata


def hash_path(path):
    if os.path.isdir(path):
        return hash_directory(path)
    else:
        with open(path, 'rb') as fp:
            return hash_file(fp)


class PersistedInputPath(Module):
    """Records or retrieves an external file in the file store.

    Because this class has the same interface for querying an existing file and
    inserting a new one, it uses Metadata instead of QueryCondition for the
    query. This means it only allows equality conditions.
    """

    _input_ports = [
            IPort('path', Path, optional=True),
            IPort('metadata', Metadata, optional=True),
            IPort('hash', PersistentHash, optional=True)]
    _output_ports = [
            OPort('path', Path)]

    def compute(self):
        localpath = self.force_get_input('path')
        hasquery = self.has_input('metadata')
        hashash = self.has_input('hash')

        file_store = get_default_store()

        if hashash:
            if localpath or hasquery:
                raise ModuleError(self,
                                  "Don't set other ports if 'hash' is set")
            h = self.get_input('hash')._hash
            self._set_result(file_store.get(h))
        elif hasquery:
            # Do the query
            metadata = self.get_input_list('metadata')
            metadata = dict(m.metadata for m in metadata)
            # Find the most recent match
            best = None
            for entry in file_store.query(metadata):
                if best is None or (KEY_TIME in entry.metadata and
                        entry[KEY_TIME] > best[KEY_TIME]):
                    best = entry
            if best is not None:
                self.check_path_type(best.filename)

            if localpath and os.path.exists(localpath.name):
                path = localpath.name
                self.check_path_type(path)
                if best is not None:
                    # Compare
                    if hash_path(path) != best['hash']:
                        # Record new version of external file
                        use_local = True
                    else:
                        # Recorded version is up to date
                        use_local = False
                else:
                    # No external file: use recorded version
                    use_local = True
                if use_local:
                    data = dict(metadata)
                    data[KEY_TYPE] = TYPE_INPUT
                    data[KEY_TIME] = datetime.strftime(datetime.utcnow(),
                                                       '%Y-%m-%d %H:%M:%S')
                    best = file_store.add(path, data)
                    self.annotate({'added_file': best['hash']})
            elif localpath:
                debug.warning("Local file does not exist: %s" % localpath)
            if best is None:
                raise ModuleError(self, "Query returned no file")
            self._set_result(best)
        else:
            raise ModuleError(self,
                              "Missing input: set either 'metadata' "
                              "(optionally with path) or hash")

    def check_path_type(self, path):
        pass

    def _set_result(self, entry):
        self.set_output('path', PathObject(entry.filename))
        # TODO : output metadata


class PersistedInputFile(PersistedInputPath):
    """Records or retrieves an external file in the file store.

    Because this class has the same interface for querying an existing file and
    inserting a new one, it uses Metadata instead of QueryCondition for the
    query. This means it only allows equality conditions.
    """

    _input_ports = [
            IPort('path', File, optional=True),
            IPort('metadata', Metadata, optional=True),
            IPort('hash', PersistentHash, optional=True)]
    _output_ports = [
            OPort('path', File)]
    _settings = ModuleSettings(configure_widget=
            'vistrails.packages.persistent_archive.widgets:SetMetadataWidget')

    def check_path_type(self, path):
        if not os.path.isfile(path):
            raise ModuleError(self, "Path is not a file")


class PersistedInputDir(PersistedInputPath):
    """Records or retrieves an external directory in the file store.

    Because this class has the same interface for querying an existing
    directory and inserting a new one, it uses Metadata instead of
    QueryCondition for the query. This means it only allows equality
    conditions.
    """

    _input_ports = [
            IPort('path', Directory,  optional=True),
            IPort('metadata', Metadata,  optional=True),
            IPort('hash', PersistentHash,  optional=True)]
    _output_ports = [
            OPort('path', File)]
    _settings = ModuleSettings(configure_widget=
            'vistrails.packages.persistent_archive.widgets:SetMetadataWidget')

    def check_path_type(self, path):
        if not os.path.isdir(path):
            raise ModuleError(self, "Path is not a directory")



#!/usr/bin/env python

# This example demonstrates how to combine a "geometric" implicit
# function with noise at different frequencies to produce the
# appearance of a landscape.

import vtk

plane = vtk.vtkPlane()

p1 = vtk.vtkPerlinNoise()
p1.SetFrequency(1, 1, 0)

p2 = vtk.vtkPerlinNoise()
p2.SetFrequency(3, 5, 0)
p2.SetPhase(0.5, 0.5, 0)

p3 = vtk.vtkPerlinNoise()
p3.SetFrequency(16, 16, 0)

sum = vtk.vtkImplicitSum()
sum.SetNormalizeByWeight(1)
sum.AddFunction(plane)
sum.AddFunction(p1, 0.2)
sum.AddFunction(p2, 0.1)
sum.AddFunction(p3, 0.02)

sample = vtk.vtkSampleFunction()
sample.SetImplicitFunction(sum)
sample.SetSampleDimensions(65, 65, 20)
sample.SetModelBounds(-1, 1, -1, 1, -0.5, 0.5)
sample.ComputeNormalsOff()
surface = vtk.vtkContourFilter()
surface.SetInputConnection(sample.GetOutputPort())
surface.SetValue(0, 0.0)

smooth = vtk.vtkPolyDataNormals()
smooth.SetInputConnection(surface.GetOutputPort())
smooth.SetFeatureAngle(90)

mapper = vtk.vtkPolyDataMapper()
mapper.SetInputConnection(smooth.GetOutputPort())
mapper.ScalarVisibilityOff()
actor = vtk.vtkActor()
actor.SetMapper(mapper)
actor.GetProperty().SetColor(0.4, 0.2, 0.1)

# Create the renderer etc.
ren = vtk.vtkRenderer()
renWin = vtk.vtkRenderWindow()
renWin.AddRenderer(ren)
iren = vtk.vtkRenderWindowInteractor()
iren.SetRenderWindow(renWin)

# Add the actors to the renderer, set the background and size
ren.AddActor(actor)
ren.SetBackground(1, 1, 1)
renWin.SetSize(500, 500)
ren.ResetCamera()
ren.GetActiveCamera().Elevation(-45)
ren.GetActiveCamera().Azimuth(10)
ren.GetActiveCamera().Dolly(1.35)
ren.ResetCameraClippingRange()

iren.Initialize()
renWin.Render()
iren.Start()



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
# search functions for entity-type objects
# notes in plain text, not html, should be fix later
from __future__ import division

import datetime
import re
import time
import unittest

from vistrails.core.query import extract_text

################################################################################

class SearchParseError(Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)

class SearchStmt(object):
    def __init__(self, content):
        self.text = content
        self.content = re.compile('.*'+content+'.*', re.MULTILINE | re.IGNORECASE)

    def match(self, entity):
        return True

    def matchModule(self, v, m):
        return True

    def run(self, v, n):
        pass

    def __call__(self):
        """Make SearchStmt behave just like a QueryObject."""
        return self

class TimeSearchStmt(SearchStmt):
    oneSecond = 1.0
    oneMinute = oneSecond * 60.0
    oneHour = oneMinute * 60.0
    oneDay = oneHour * 24.0
    oneWeek = oneDay * 7.0
    oneMonth = oneDay * 31.0 # wrong, I know
    oneYear = oneDay * 365.0 # wrong, I know
    amounts = {'seconds': oneSecond,
               'minutes': oneMinute,
               'hours': oneHour,
               'days': oneDay,
               'weeks': oneWeek,
               'months': oneMonth,
               'years': oneYear}
    months = {'january': 1,
              'february': 2,
              'march': 3,
              'april': 4,
              'may': 5,
              'june': 6,
              'july': 7,
              'august': 8,
              'september': 9,
              'october': 10,
              'november': 11,
              'december': 12}
    
    dateEntry = r'([^\,\/\: ]+)'
    timeEntry = r'(\d?\d?)'
    dateSep = r' *[\,\/\- ] *'
    timeSep = r' *: *'
    sep = r' *'
    start = r'^ *'
    finish = r' *$'
    twoEntryDate = (dateEntry+
                    dateSep+
                    dateEntry)
    threeEntryDate = (dateEntry+
                      dateSep+
                      dateEntry+
                      dateSep+
                      dateEntry)
    twoEntryTime = (timeEntry+
                    timeSep+
                    timeEntry)
    threeEntryTime = (timeEntry+
                      timeSep+
                      timeEntry+
                      timeSep+
                      timeEntry)

    dateRE = [re.compile((start+
                          twoEntryDate+
                          finish)), # Mar 12   Mar, 12    
              re.compile((start+
                          threeEntryDate+
                          finish)), # Mar, 12, 2006    2006 Mar 12     etc
              re.compile((start+
                          twoEntryTime+
                          finish)),
              re.compile((start+
                          threeEntryTime+
                          finish)),
              re.compile((start+
                          twoEntryDate+
                          sep+
                          twoEntryTime+
                          finish)),
              re.compile((start+
                          twoEntryDate+
                          sep+
                          threeEntryTime+
                          finish)),
              re.compile((start+
                          threeEntryDate+
                          sep+
                          twoEntryTime+
                          finish)),
              re.compile((start+
                          threeEntryDate+
                          sep+
                          threeEntryTime+
                          finish)),
              re.compile((start+
                          twoEntryTime+
                          sep+
                          twoEntryDate+
                          finish)),
              re.compile((start+
                          twoEntryTime+
                          sep+
                          threeEntryDate+
                          finish)),
              re.compile((start+
                          threeEntryTime+
                          sep+
                          twoEntryDate+
                          finish)),
              re.compile((start+
                          threeEntryTime+
                          sep+
                          threeEntryDate+
                          finish))]
    
    def __init__(self, date):
        self.date = self.parseDate(date)

    def parseDate(self, dateStr):
        def parseAgo(s):
            [amount, unit] = s.split(' ')
            try:
                amount = float(amount)
            except ValueError:
                raise SearchParseError("Expected a number, got %s" % amount)
            if amount <= 0:
                raise SearchParseError("Expected a positive number, got %s" % amount)
            unitRe = re.compile('^'+unit)
            keys = [k
                    for k in TimeSearchStmt.amounts.keys()
                    if unitRe.match(k)]
            if len(keys) == 0:
                raise SearchParseError("Time unit unknown: %s" % unit)
            elif len(keys) > 1:
                raise SearchParseError("Time unit ambiguous: %s matches %s" % (unit, keys))
            return round(time.time()) - TimeSearchStmt.amounts[keys[0]] * amount

        def guessDate(unknownEntries, year=None):
            def guessStrMonth(s):
                monthRe = re.compile('^'+s)
                keys = [k
                        for k in TimeSearchStmt.months.keys()
                        if monthRe.match(k)]
                if len(keys) == 0:
                    raise SearchParseError("Unknown month: %s" % s)
                elif len(keys) > 1:
                    raise SearchParseError("Ambiguous month: %s matches %s" % (s, keys))
                return TimeSearchStmt.months[keys[0]]
            if not year:
                m = None
                # First heuristic: if month comes first, then year comes last
                try:
                    e0 = int(unknownEntries[0])
                except ValueError:
                    m = guessStrMonth(unknownEntries[0])
                    try:
                        d = int(unknownEntries[1])
                    except ValueError:
                        raise SearchParseError("Expected day, got %s" % unknownEntries[1])
                    try:
                        y = int(unknownEntries[2])
                    except ValueError:
                        raise SearchParseError("Expected year, got %s" % unknownEntries[2])
                    return (y, m, d)
                # Second heuristic: if month comes last, then year comes first
                try:
                    e2 = int(unknownEntries[2])
                except ValueError:
                    m = guessStrMonth(unknownEntries[2])
                    try:
                        d = int(unknownEntries[1])
                    except ValueError:
                        raise SearchParseError("Expected day, got %s" % unknownEntries[1])
                    try:
                        y = int(unknownEntries[0])
                    except ValueError:
                        raise SearchParseError("Expected year, got %s" % unknownEntries[0])
                    return (y, m, d)
                # If month is the middle one, decide day and year by size
                # (year is largest, hopefully year was entered using 4 digits)
                try:
                    e1 = int(unknownEntries[1])
                except ValueError:
                    m = guessStrMonth(unknownEntries[1])
                    try:
                        d = int(unknownEntries[2])
                    except ValueError:
                        raise SearchParseError("Expected day or year, got %s" % unknownEntries[2])
                    try:
                        y = int(unknownEntries[0])
                    except ValueError:
                        raise SearchParseError("Expected year or year, got %s" % unknownEntries[0])
                    return (max(y,d), m, min(y, d))
                lst = [(e0,0),(e1,1),(e2,2)]
                lst.sort()
                return guessDate([str(lst[0][0]),
                                  str(lst[1][0])],
                                 year=e2)
            # We know year, decide month using similar heuristics - try string month first,
            # then decide which is possible
            try:
                e0 = int(unknownEntries[0])
            except ValueError:
                m = guessStrMonth(unknownEntries[0])
                try:
                    d = int(unknownEntries[1])
                except ValueError:
                    raise SearchParseError("Expected day, got %s" % unknownEntries[1])
                return (year, m, d)
            try:
                e1 = int(unknownEntries[1])
            except ValueError:
                m = guessStrMonth(unknownEntries[1])
                try:
                    d = int(unknownEntries[0])
                except ValueError:
                    raise SearchParseError("Expected day, got %s" % unknownEntries[0])
                return (year, m, d)
            if e0 > 12:
                return (year, e1, e0)
            else:
                return (year, e0, e1)

        dateStr = dateStr.lower().lstrip().rstrip()
        if dateStr.endswith(" ago"):
            return parseAgo(dateStr[:-4])
        if dateStr == "yesterday":
            lst = list(time.localtime(round(time.time()) - TimeSearchStmt.oneDay))
            # Reset hour, minute, second
            lst[3] = 0
            lst[4] = 0
            lst[5] = 0
            return time.mktime(lst)
        if dateStr == "today":
            lst = list(time.localtime())
            # Reset hour, minute, second
            lst[3] = 0
            lst[4] = 0
            lst[5] = 0
            return time.mktime(lst)
        if dateStr.startswith("this "):
            rest = dateStr[5:]
            lst = list(time.localtime(round(time.time())))
            if rest == "minute":
                lst[5] = 0
            elif rest == "hour":
                lst[5] = 0
                lst[4] = 0
            elif rest == "day":
                lst[5] = 0
                lst[4] = 0
                lst[3] = 0
            elif rest == "week": # weeks start on monday
                lst[5]  = 0
                lst[4]  = 0
                lst[3]  = 0
                # This hack saves me the hassle of computing negative days, months, etc
                lst = list(time.localtime(time.mktime(lst) - TimeSearchStmt.oneDay * lst[6]))
            elif rest == "month":
                lst[5]  = 0
                lst[4]  = 0
                lst[3]  = 0
                lst[2]  = 1
            elif rest == "year":
                lst[5]  = 0
                lst[4]  = 0
                lst[3]  = 0
                lst[2]  = 1
                lst[1]  = 1
            return time.mktime(lst)
                
        result = [x.match(dateStr) for x in TimeSearchStmt.dateRE]
        this = list(time.localtime())
        def setTwoDate(g):
            d = guessDate(g, year=this[0])
            this[0] = d[0]
            this[1] = d[1]
            this[2] = d[2]
        def setThreeDate(g):
            d = guessDate(g)
            this[0] = d[0]
            this[1] = d[1]
            this[2] = d[2]
        def setTwoTime(g):
            this[3] = int(g[0])
            this[4] = int(g[1])
            this[5] = 0
        def setThreeTime(g):
            this[3] = int(g[0])
            this[4] = int(g[1])
            this[5] = int(g[2])
        if result[0]:
            setTwoDate(result[0].groups())
            setTwoTime([0,0])
        elif result[1]:
            setThreeDate(result[1].groups())
            setTwoTime([0,0])
        elif result[2]:
            setTwoTime(result[2].groups())
        elif result[3]:
            setThreeTime(result[3].groups())
        elif result[4]:
            g = result[4].groups()
            setTwoDate([g[0], g[1]])
            setTwoTime([g[2], g[3]])
        elif result[5]:
            g = result[5].groups()
            setTwoDate([g[0], g[1]])
            setThreeTime([g[2], g[3], g[4]])
        elif result[6]:
            g = result[6].groups()
            setThreeDate([g[0], g[1], g[2]])
            setTwoTime([g[3], g[4]])
        elif result[7]:
            g = result[7].groups()
            setThreeDate([g[0], g[1], g[2]])
            setThreeTime([g[3], g[4], g[5]])
        elif result[8]:
            g = result[8].groups()
            setTwoTime([g[0], g[1]])
            setTwoDate([g[2], g[3]])
        elif result[9]:
            g = result[9].groups()
            setTwoTime([g[0], g[1]])
            setThreeDate([g[2], g[3], g[4]])
        elif result[10]:
            g = result[10].groups()
            setThreeTime([g[0], g[1], g[2]])
            setTwoDate([g[3], g[4]])
        elif result[11]:
            g = result[11].groups()
            setThreeTime([g[0], g[1], g[2]])
            setThreeDate([g[3], g[4],g[5]])
        else:
            raise SearchParseError("Expected a date, got '%s'" % dateStr)
        return time.mktime(this)
        
class BeforeSearchStmt(TimeSearchStmt):
    def match(self, entity):
        if not entity.mod_time:
            return False
        t = time.mktime(entity.mod_time)
        return t <= self.date

class AfterSearchStmt(TimeSearchStmt):
    def match(self, entity):
        if not entity.mod_time:
            return False
        t = time.mktime(entity.mod_time)
        return t >= self.date

class UserSearchStmt(SearchStmt):
    def match(self, entity):
        if not entity.user:
            return False
        return self.content.match(entity.user)

class NotesSearchStmt(SearchStmt):
    def match(self, entity):
        if entity.description:
            plainNotes = extract_text(entity.description)
            return self.content.search(plainNotes)
        return False

class NameSearchStmt(SearchStmt):
    def match(self, entity):
        return self.content.match(entity.name)

class AndSearchStmt(SearchStmt):
    def __init__(self, lst):
        self.matchList = lst
    def match(self, entity):
        for s in self.matchList:
            if not s.match(entity):
                return False
        return True

class OrSearchStmt(SearchStmt):
    def __init__(self, lst):
        self.matchList = lst
    def match(self, entity):
        for s in self.matchList:
            if s.match(entity):
                return True
        return False

class NotSearchStmt(SearchStmt):
    def __init__(self, stmt):
        self.stmt = stmt
    def match(self, entity):
        return not self.stmt.match(entity)

class TrueSearch(SearchStmt):
    def __init__(self):
        pass
    def match(self, entity):
        return True

################################################################################

class SearchCompiler(object):
    SEPARATOR = -1
    def __init__(self, searchStr):
        self.searchStmt = self.compile(searchStr)
    def compile(self, searchStr):
        lst = []
        t1 = searchStr.split(' ')
        while t1:
            tok = t1[0]
            cmd = tok.split(':')
            if not SearchCompiler.dispatch.has_key(cmd[0]):
                fun = SearchCompiler.parseAny
            else:
                fun = SearchCompiler.dispatch[cmd[0]]
            if len(cmd) > 1:
                [search, rest] = fun(self, cmd[1:] + t1[1:])
            else:
                [search, rest] = fun(self, t1)
            lst.append(search)
            t1 = rest
        return AndSearchStmt(lst)
    def parseUser(self, tokStream):
        if len(tokStream) == 0:
            raise SearchParseError('Expected token, got end of search')
        return (UserSearchStmt(tokStream[0]), tokStream[1:])
    def parseAny(self, tokStream):
        if len(tokStream) == 0:
            raise SearchParseError('Expected token, got end of search')
        tok = tokStream[0]
        return (OrSearchStmt([UserSearchStmt(tok),
                              NotesSearchStmt(tok),
                              NameSearchStmt(tok)]), tokStream[1:])
    def parseNotes(self, tokStream):
        if len(tokStream) == 0:
            raise SearchParseError('Expected token, got end of search')
        lst = []
        while len(tokStream):
            tok = tokStream[0]
            if ':' in tok:
                return (AndSearchStmt(lst), tokStream)
            lst.append(NotesSearchStmt(tok))
            tokStream = tokStream[1:]
        return (AndSearchStmt(lst), [])
    def parseName(self, tokStream):
        if len(tokStream) == 0:
            raise SearchParseError('Expected token, got end of search')
        lst = []
        while len(tokStream):
            tok = tokStream[0]
            if ':' in tok:
                return (AndSearchStmt(lst), tokStream)
            lst.append(NameSearchStmt(tok))
            tokStream = tokStream[1:]
        return (AndSearchStmt(lst), [])
    def parseBefore(self, tokStream):
        old_tokstream = tokStream
        try:
            if len(tokStream) == 0:
                raise SearchParseError('Expected token, got end of search')
            lst = []
            while len(tokStream):
                tok = tokStream[0]
                # ugly, special case times
                if (':' in tok and
                    not TimeSearchStmt.dateRE[2].match(tok) and
                    not TimeSearchStmt.dateRE[3].match(tok)):
                    return (BeforeSearchStmt(" ".join(lst)), tokStream)
                lst.append(tok)
                tokStream = tokStream[1:]
            return (BeforeSearchStmt(" ".join(lst)), [])
        except SearchParseError, e:
            if 'Expected a date' in e.args[0]:
                try:
                    return self.parseAny(old_tokstream)
                except SearchParseError, e2:
                    print "Another exception...", e2.args[0]
                    raise e
            else:
                raise
            
    def parseAfter(self, tokStream):
        try:
            if len(tokStream) == 0:
                raise SearchParseError('Expected token, got end of search')
            lst = []
            while len(tokStream):
                tok = tokStream[0]
                # ugly, special case times
                if (':' in tok and
                    not TimeSearchStmt.dateRE[2].match(tok) and
                    not TimeSearchStmt.dateRE[3].match(tok)):
                    return (AfterSearchStmt(" ".join(lst)), tokStream)
                lst.append(tok)
                tokStream = tokStream[1:]
            return (AfterSearchStmt(" ".join(lst)), [])
        except SearchParseError, e:
            if 'Expected a date' in e.args[0]:
                try:
                    return self.parseAny(['after'] + tokStream)
                except SearchParseError, e2:
                    print "Another exception...", e2.args[0]
                    raise e
            else:
                raise

    dispatch = {'user': parseUser,
                'notes': parseNotes,
                'before': parseBefore,
                'after': parseAfter,
                'name': parseName,
                'any': parseAny}
                
            

################################################################################


class TestSearch(unittest.TestCase):
    def test1(self):
        self.assertEquals((TimeSearchStmt('1 day ago').date -
                           TimeSearchStmt('2 days ago').date), TimeSearchStmt.oneDay)
    def test2(self):
        self.assertEquals((TimeSearchStmt('12 mar 2006').date -
                           TimeSearchStmt('11 mar 2006').date), TimeSearchStmt.oneDay)
    def test3(self):
        # This will fail if year flips during execution. Oh well :)
        yr = datetime.datetime.today().year
        self.assertEquals((TimeSearchStmt('12 mar').date -
                           TimeSearchStmt('12 mar %d' % yr).date), 0.0)
    def test4(self):
        # This will fail if year flips during execution. Oh well :)
        yr = datetime.datetime.today().year
        self.assertEquals((TimeSearchStmt('mar 12').date -
                           TimeSearchStmt('12 mar %d' % yr).date), 0.0)
    def test5(self):
        yr = datetime.datetime.today().year
        self.assertEquals((TimeSearchStmt('03 15').date -
                           TimeSearchStmt('15 mar %d' % yr).date), 0.0)
    def test6(self):
        self.assertEquals((TimeSearchStmt('03/15/2006').date -
                           TimeSearchStmt('15 mar 2006').date), 0.0)
    def test7(self):
        self.assertEquals((TimeSearchStmt('1 day ago').date -
                           TimeSearchStmt('24 hours ago').date), 0.0)
    def test8(self):
        self.assertEquals((TimeSearchStmt('1 hour ago').date -
                           TimeSearchStmt('60 minutes ago').date), 0.0)
    def test9(self):
        self.assertEquals((TimeSearchStmt('1 minute ago').date -
                           TimeSearchStmt('60 seconds ago').date), 0.0)
    def test10(self):
        self.assertEquals((TimeSearchStmt('1 week ago').date -
                           TimeSearchStmt('7 days ago').date), 0.0)
    def test11(self):
        self.assertEquals((TimeSearchStmt('1 month ago').date -
                           TimeSearchStmt('31 days ago').date), 0.0)
    def test12(self):
        self.assertEquals(TimeSearchStmt('12 mar 2007 21:00:00').date,
                          TimeSearchStmt('21:00:00 12 mar 2007').date)
    def test13(self):
        # This will fail if year flips during execution. Oh well :)
        yr = datetime.datetime.today().year
        self.assertEquals(TimeSearchStmt('12 mar %d 21:00' % yr).date,
                          TimeSearchStmt('21:00:00 12 mar').date)
    def test14(self):
        self.assertEquals(TimeSearchStmt('13 apr 2006 21:00').date,
                          TimeSearchStmt('04/13/2006 21:00:00').date)
    def test15(self):
        import vistrails.core.vistrail
        from vistrails.core.db.locator import XMLFileLocator
        import vistrails.core.system
        v = XMLFileLocator(vistrails.core.system.vistrails_root_directory() +
                           '/tests/resources/dummy.xml').load()
        # FIXME: Add notes to this.
#         self.assertTrue(NotesSearchStmt('mapper').match(v.actionMap[36]))
#         self.assertFalse(NotesSearchStmt('-qt-block-indent').match(v.actionMap[36]))

    # test16 and 17 now pass.
    #     def test16(self):
    #         self.assertRaises(SearchParseError, lambda *args: SearchCompiler('before:'))
    #     def test17(self):
    #         self.assertRaises(SearchParseError, lambda *args: SearchCompiler('after:yesterday before:lalala'))
    def test18(self):
        self.assertEquals(TimeSearchStmt('   13 apr 2006  ').date,
                          TimeSearchStmt(' 13 apr 2006   ').date)
    def test19(self):
        self.assertEquals(SearchCompiler('before:13 apr 2006 12:34:56').searchStmt.matchList[0].date,
                          BeforeSearchStmt('13 apr 2006 12:34:56').date)
    def test20(self):
        self.assertEquals(SearchCompiler('after:yesterday').searchStmt.matchList[0].date,
                          SearchCompiler('before:yesterday').searchStmt.matchList[0].date)
    def test21(self):
        self.assertEquals(SearchCompiler('after:today').searchStmt.matchList[0].date,
                          SearchCompiler('before:today').searchStmt.matchList[0].date)
    def test22(self):
        self.assertEquals(SearchCompiler('before:today').searchStmt.matchList[0].date,
                          SearchCompiler('before:this day').searchStmt.matchList[0].date)
    def test23(self):
        t = time.localtime()
        import vistrails.core.utils
        inv = vistrails.core.utils.invert(TimeSearchStmt.months)
        m = inv[t[1]]
        self.assertEquals(SearchCompiler('after:%s %s %s' % (t[0], m, t[2])).searchStmt.matchList[0].date,
                          SearchCompiler('after:today').searchStmt.matchList[0].date)
    def test24(self):
        # Test compiling these searches
        SearchCompiler('before')
        SearchCompiler('after')

if __name__ == '__main__':
    unittest.main()



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
from __future__ import division

from PyQt4 import QtCore, QtGui
from PyQt4.QtCore import pyqtSignal

from vistrails.core import debug
from vistrails.gui.mashups.mashups_widgets import QAliasNumericStepperWidget, \
    QAliasSliderWidget, QDropDownWidget
from vistrails.gui.utils import show_warning, TestVisTrailsGUI

from vistrails.packages.spreadsheet.spreadsheet_controller import \
    spreadsheetController


class QMashupAppMainWindow(QtGui.QMainWindow):
    #signals
    appWasClosed = pyqtSignal(QtGui.QMainWindow)

    def __init__(self, parent=None, vistrail_view=None, dumpcells=False,
                 controller=None, version=-1):
        """ QMashupAppMainWindow()
        Initialize an app window from a mashup.

        """
        # Constructing the main widget
        QtGui.QMainWindow.__init__(self, parent)
        self.vtkCells = []
        self.setStatusBar(QtGui.QStatusBar(self))

        # Central widget
        centralWidget = QtGui.QWidget()
        self.mainLayout = QtGui.QVBoxLayout()
        self.mainLayout.setMargin(0)
        self.mainLayout.setSpacing(5)
        centralWidget.setLayout(self.mainLayout)
        self.setCentralWidget(centralWidget)
        self.numberOfCells = 0
        self.is_executing = False
        self.sequenceOption = False
        self.steps = []
        self.isLooping = False
        #self.resize(100,100)
        self.dumpcells = dumpcells
        self.view = vistrail_view
        if controller:
            self.controller = controller
            self.mshptrail = controller.mshptrail
            if version == -1:
                self.currentMashup = self.controller.currentMashup
            else:
                self.currentMashup = self.mshptrail.getMashup(version)
            self.setWindowTitle('%s Mashup'%self.controller.getMashupName(version))
        else:
            self.setWindowTitle('Mashup')

        # Assign "hidden" shortcut
        self.editingModeAct = QtGui.QAction("Chang&e Layout",
                                 self, shortcut="Ctrl+E",
                                 statusTip="Change the layout of the widgets",
                                 triggered=self.toggleEditingMode)
        #self.editingModeShortcut = QtGui.QShortcut(QtGui.QKeySequence('Ctrl+E'), self)
        #self.connect(self.editingModeShortcut, QtCore.SIGNAL('activated()'),
        #             self.toggleEditingMode)
        self.editing = False
        # Constructing alias controls
        self.controlDocks = []
        # Show here to make sure XDisplay info is correct (for VTKCell)
        self.show()

        spreadsheetController.setEchoMode(True)
        #will run to get Spreadsheet Cell events
        (cellEvents, errors) = self.runAndGetCellEvents(useDefaultValues=True)
        if cellEvents:
            self.numberOfCells = len(cellEvents)
            self.initCells(cellEvents)
        if len(errors) > 0:
            show_warning("VisTrails::Mashup Preview",
                         "There was a problem executing the pipeline: %s." %
                         errors)
        # Construct the controllers for aliases
        self.controlDocks = {}
        self.cellControls = {}
        self.aliasWidgets = {}
        self.initControls()

        if self.currentMashup.layout is not None:
            self.restoreState(QtCore.QByteArray.fromPercentEncoding(
                                QtCore.QByteArray(self.currentMashup.layout)))

        if self.currentMashup.geometry is not None:
            self.restoreGeometry(QtCore.QByteArray.fromPercentEncoding(
                              QtCore.QByteArray(self.currentMashup.geometry)))
        else:
            self.resize(self.sizeHint())

        # Constructing buttons
        buttonDock = QCustomDockWidget('Control Buttons', self)
        buttonWidget = QtGui.QWidget(buttonDock)
        buttonWidget.setSizePolicy(QtGui.QSizePolicy.Preferred,
                                   QtGui.QSizePolicy.Preferred)
        buttonLayout = QtGui.QGridLayout()
        buttonWidget.setLayout(buttonLayout)
        buttonLayout.setMargin(5)
        self.cb_auto_update = QtGui.QCheckBox("Turn on auto-update",
                                              self.centralWidget())
        self.cb_auto_update.setChecked(False)
        self.cb_loop_sequence = QtGui.QCheckBox("Render all steps in '%s'" % \
           (self.sequenceOption.alias.name if self.sequenceOption else 'None'),
            self.centralWidget())
        self.cb_loop_sequence.setChecked(False)
        self.cb_loop_sequence.setVisible(self.sequenceOption is not False)
        self.cb_loop_sequence.setToolTip(
                        "Render each step of this stepper for fast switching")
        self.cb_loop_int = QtGui.QCheckBox("Interactive Steps",
                                           self.centralWidget())
        self.cb_loop_int.setChecked(False)
        self.cb_loop_int.setVisible(False)
        self.cb_loop_int.setToolTip(
                 "Show complete result of each step instead of static images")
        self.cb_loop_sequence.clicked.connect(self.cb_loop_int.setVisible)
        self.cb_keep_camera = QtGui.QCheckBox("Keep camera position",
                                              self.centralWidget())
        self.cb_keep_camera.setChecked(True)
        self.connect(self.cb_auto_update,
                     QtCore.SIGNAL("stateChanged(int)"),
                     self.auto_update_changed)
        self.connect(self.cb_loop_int,
                     QtCore.SIGNAL("stateChanged(int)"),
                     self.loop_int_changed)
        self.loopButton = QtGui.QPushButton("&Loop", self.centralWidget())
        self.loopButton.setToolTip("Loop automatically through steps")
        self.loopButton.setCheckable(True)
        self.loopButton.setVisible(self.sequenceOption is not False)
        self.updateButton = QtGui.QPushButton("&Update", self.centralWidget())
        if self.dumpcells:
            self.quitButton = QtGui.QPushButton("&Save", self.centralWidget())
            self.connect(self.quitButton,
                         QtCore.SIGNAL('clicked(bool)'),
                         self.saveAndExport)
        else:
            self.quitButton = QtGui.QPushButton("&Quit", self.centralWidget())
            self.connect(self.quitButton,
                         QtCore.SIGNAL('clicked(bool)'),
                         self.close)
        buttonLayout.setColumnStretch(0, 1)
        if self.sequenceOption:
            sequenceLayout = QtGui.QHBoxLayout()
            sequenceLayout.setMargin(5)
            sequenceLayout.addWidget(self.cb_loop_int)
            sequenceLayout.addWidget(self.cb_loop_sequence)
            buttonLayout.addLayout(sequenceLayout, 0, 0, QtCore.Qt.AlignRight)
        buttonLayout.addWidget(self.cb_auto_update, 0, 1, QtCore.Qt.AlignLeft)
        buttonLayout.addWidget(self.cb_keep_camera, 0, 2, 1, 2, QtCore.Qt.AlignLeft)
        if self.sequenceOption:
            buttonLayout.addWidget(self.loopButton, 1, 1, QtCore.Qt.AlignRight)
            self.loopButton.setEnabled(False)
        buttonLayout.addWidget(self.updateButton, 1, 2, QtCore.Qt.AlignRight)
        buttonLayout.addWidget(self.quitButton, 1, 3, QtCore.Qt.AlignRight)
        self.connect(self.updateButton,
                     QtCore.SIGNAL('clicked(bool)'),
                     self.updateButtonClick)
        if self.sequenceOption:
            self.connect(self.loopButton,
                         QtCore.SIGNAL('clicked(bool)'),
                         self.loopButtonClick)
        buttonDock.setWidget(buttonWidget)
        self.addDockWidget(QtCore.Qt.BottomDockWidgetArea, buttonDock)
        self.controlDocks["__buttons__"] = buttonDock

        self.saveAllAct = QtGui.QAction("S&ave Combined", self,
                                        shortcut=QtGui.QKeySequence.SelectAll,
                                        statusTip="Save combined images to disk",
                                        triggered=self.saveAllEvent)
        self.saveAct = QtGui.QAction("&Save Each", self,
                                     shortcut=QtGui.QKeySequence.Save,
                                     statusTip="Save separate images to disk",
                                     triggered=self.saveEventAction)
        self.showBuilderAct = QtGui.QAction("VisTrails Main Window", self,
                                            statusTip="Show VisTrails Main Window",
                                            triggered=self.showBuilderWindow)
        self.createMenus()
        self.lastExportPath = ''

    def createMenus(self):
        self.fileMenu = self.menuBar().addMenu("&File")
        self.fileMenu.addAction(self.saveAct)
        self.fileMenu.addAction(self.saveAllAct)

        self.viewMenu = self.menuBar().addMenu("&View")
        self.viewMenu.addAction(self.editingModeAct)

        self.windowMenu = self.menuBar().addMenu("&Window")
        self.windowMenu.addAction(self.showBuilderAct)

    def runAndGetCellEvents(self, useDefaultValues=False):
        spreadsheetController.setEchoMode(True)
        #will run to get Spreadsheet Cell events
        cellEvents = []
        errors = []
        try:
            (res, errors) = self.run(useDefaultValues)
            if res:
                cellEvents = spreadsheetController.getEchoCellEvents()
        except Exception, e:
            debug.unexpected_exception(e)
            print "Executing pipeline failed:", debug.format_exc()
        finally:
            spreadsheetController.setEchoMode(False)

        return (cellEvents, errors)

    def updateCells(self, info=None):
        # check if we should create a sequence
        if self.cb_loop_sequence.isChecked():
            return self.updateCellsLoop(info)
        self.is_executing = True
        (cellEvents, errors) = self.runAndGetCellEvents()
        self.is_executing = False
        if len(cellEvents) != self.numberOfCells:
            raise RuntimeError(
                    "The number of cells has changed (unexpectedly) "
                    "(%d vs. %d)!\n"
                    "Pipeline results: %s" % (len(cellEvents),
                                              self.numberOfCells,
                                              errors))
        #self.SaveCamera()
        for i in xrange(self.numberOfCells):
            camera = []
            if (hasattr(self.cellWidgets[i],"getRendererList") and
                self.cb_keep_camera.isChecked()):
                for ren in self.cellWidgets[i].getRendererList():
                    camera.append(ren.GetActiveCamera())
                self.cellWidgets[i].updateContents(cellEvents[i].inputPorts, camera)
                #self.cellWidgets[i].updateContents(cellEvents[i].inputPorts)
            else:
                self.cellWidgets[i].updateContents(cellEvents[i].inputPorts)

    def updateCellsLoop(self, info=None):
        """ Run workflow for each step in the loop sequence and collect results.
        """
        interactive = self.cb_loop_int.isChecked()
        slider = self.sequenceOption.value
        if info and info[1][0] == slider:
            # User is moving the slider, so we use the existing result
            if interactive:
                if slider.value() < len(self.steps):
                    self.updateRenderedCells(slider.value())
            else:
                for i in xrange(self.numberOfCells):
                    self.cellWidgets[i].setPlayerFrame(slider.value())
            return

        if not interactive:
            for i in xrange(self.numberOfCells):
                self.cellWidgets[i].clearHistory()

        self.is_executing = True
        self.steps = []
        old_value = slider.value()
        value = slider.minimum()
        slider.setValue(value)

        while True:
            (cellEvents, errors) = self.runAndGetCellEvents()
            if len(cellEvents) != self.numberOfCells:
                raise RuntimeError(
                        "The number of cells has changed (unexpectedly) "
                        "(%d vs. %d)!\n"
                        "Pipeline results: %s" % (len(cellEvents),
                                                  self.numberOfCells,
                                                  errors))
            if interactive:
                self.steps.append([])
            else:
                self.steps = [[]]
            for i in xrange(self.numberOfCells):
                self.steps[-1].append(cellEvents[i].inputPorts)

            # show the result
            self.updateRenderedCells(value if interactive else 0)
            self.is_executing = True

            if value >= slider.maximum():
                break
            value += slider.singleStep()
            slider.setValue(value)
        self.is_executing = False
        slider.setValue(old_value)
        self.loopButton.setEnabled(True)

    def updateRenderedCells(self, value):
        """ Show the cell specified by slider info
        """
        self.is_executing = True
        for i in xrange(self.numberOfCells):
            camera = []
            if (hasattr(self.cellWidgets[i],"getRendererList") and
                self.cb_keep_camera.isChecked()):
                for ren in self.cellWidgets[i].getRendererList():
                    camera.append(ren.GetActiveCamera())
                self.cellWidgets[i].updateContents(self.steps[value][i], camera)
            else:
                self.cellWidgets[i].updateContents(self.steps[value][i])
        self.is_executing = False

    def updateButtonClick(self):
        self.updateButton.setEnabled(False)
        try:
            self.updateCells()
        finally:
            self.updateButton.setEnabled(True)

    def loopButtonClick(self, toggled):
        self.updateButton.setEnabled(not toggled)
        self.cb_loop_int.setEnabled(not toggled)
        self.cb_loop_sequence.setEnabled(not toggled)
        self.cb_auto_update.setEnabled(not toggled)
        if self.cb_loop_int.isChecked():
            if toggled:
                if self.isLooping:
                    self.killTimer(self.isLooping)
                self.isLooping = self.startTimer(200)
            elif self.isLooping:
                self.killTimer(self.isLooping)
                self.isLooping = None
        else:
            for cell in self.cellWidgets:
                if toggled:
                    cell.startPlayer()
                else:
                    cell.stopPlayer()


    def timerEvent(self, event):
        if self.steps:
            stepper = self.sequenceOption.value
            if stepper.value() == stepper.maximum():
                stepper.setValue(stepper.minimum())
            else:
                stepper.setValue(stepper.value() + stepper.singleStep())
            if stepper.value() >= len(self.steps):
                self.loopButton.setChecked(False)
                return

            for i in xrange(self.numberOfCells):
                self.cellWidgets[i].setAnimationEnabled(False)

            self.updateRenderedCells(stepper.value())

            for i in xrange(self.numberOfCells):
                self.cellWidgets[i].setAnimationEnabled(True)

    def toggleEditingMode(self):
        if len(self.controlDocks) > 0:
            for dock in self.controlDocks.itervalues():
                dock.toggleTitleBar()
            self.editing = not self.editing
        if not self.editing:
            self.saveSettings()

    def saveSettings(self):
        layout = self.saveState().toPercentEncoding()
        geom = self.saveGeometry().toPercentEncoding()

        self.currentMashup.layout = layout
        self.currentMashup.geometry = geom

        self.controller.setChanged(True)

        #self.controller.writeMashuptrail()

    def closeEvent(self, event):
        self.saveSettings()
        self.appWasClosed.emit(self)
        event.accept()

    def auto_update_changed(self, state):
        if state == QtCore.Qt.Unchecked:
            self.updateButton.setEnabled(True)
            self.cb_loop_int.setEnabled(True)
            self.cb_loop_sequence.setEnabled(True)
            self.loopButton.setEnabled(True)
            if self.cb_loop_sequence.isChecked() and not self.cb_loop_int.isChecked():
                for i in xrange(self.numberOfCells):
                    cell = self.cellWidgets[i]
                    cell._player.hide()
                    cell.show()

        elif state == QtCore.Qt.Checked:
            self.updateButton.setEnabled(False)
            self.cb_loop_int.setEnabled(False)
            self.cb_loop_sequence.setEnabled(False)
            self.loopButton.setEnabled(False)
            if self.cb_loop_sequence.isChecked() and not self.cb_loop_int.isChecked():
                for i in xrange(self.numberOfCells):
                    cell = self.cellWidgets[i]
                    cell._player.setParent(cell.parent())
                    cell._player.setGeometry(cell.geometry())
                    self.cellWidgets[i].setPlayerFrame(self.sequenceOption.value.value())
                    cell._player.raise_()
                    cell._player.show()
                    cell.hide()

    def loop_int_changed(self, state):
        self.loopButton.setEnabled(False)

    def saveAll(self):
        for w in self.widgets:
            w.saveAll(self.dumpcells)

    def saveEach(self):
        for w in self.widgets:
            w.saveEach(self.dumpcells, self.frameNo)

    def saveEventAction(self, checked):
        self.saveEvent()

    def saveEvent(self, folder=None):
        if folder is None:
            folder = QtGui.QFileDialog.getExistingDirectory(self,
                                                        "Save images to...",
                                                        self.lastExportPath,
                                                        QtGui.QFileDialog.ShowDirsOnly)
        if folder:
            self.dumpcells = str(folder)
            self.saveEach()
            self.lastExportPath = str(folder)

    def saveAllEvent(self, folder=None):
        if folder is None:
            folder = QtGui.QFileDialog.getExistingDirectory(self,
                                                        "Save images to...",
                                                        self.lastExportPath,
                                                        QtGui.QFileDialog.ShowDirsOnly)
        if folder:
            self.dumpcells = str(folder)
            self.saveAll()

    def saveAndExport(self, clicked=True):
        self.saveAll()

    def initCells(self, cellEvents):
        cellLayout = QtGui.QHBoxLayout()
        self.mainLayout.addLayout(cellLayout, self.numberOfCells * 2)
        self.cellWidgets = []
        vtkCells = []
        for event in cellEvents:
            cellWidget = event.cellType(self.centralWidget())
            if event.cellType.__name__ == 'QVTKWidget':
                vtkCells.append(cellWidget)
            cellWidget.show()
            cellWidget.setSizePolicy(QtGui.QSizePolicy.Expanding,
                                      QtGui.QSizePolicy.Expanding)
            cellWidget.setMinimumSize(300, 100)
            cellLayout.addWidget(cellWidget)
            self.cellWidgets.append(cellWidget)
            cellWidget.updateContents(event.inputPorts)
        def getSelectedCellWidgets():
            return vtkCells
        for cellWidget in self.vtkCells:
            cellWidget.getSelectedCellWidgets = getSelectedCellWidgets

    def initControls(self):
        if len(self.currentMashup.alias_list) == 0:
            return

        #Constructing alias controls
        self.controlDocks = {}
        self.cellControls = {}
        self.toolbuttons = {}

        row = 0
        for alias in self.currentMashup.alias_list:
            dock = QCustomDockWidget(alias.name, #"Control for '%s'" % aliasName,
                                      self)
            vtparam = self.controller.getVistrailParam(alias)

            if alias.component.widget == 'slider':
                aliasWidget = QAliasSliderWidget(alias, vtparam, dock)
                # enables looping of
                if alias.component.seq:
                    self.sequenceOption = aliasWidget

            elif alias.component.widget == 'numericstepper':
                aliasWidget = QAliasNumericStepperWidget(alias, vtparam, dock)
            else:
                aliasWidget = QDropDownWidget(alias, vtparam, dock)

            aliasWidget.setSizePolicy(QtGui.QSizePolicy.Preferred,
                                          QtGui.QSizePolicy.Maximum)
            self.connect(aliasWidget,
                             QtCore.SIGNAL("contentsChanged"),
                             self.widget_changed)

            dock.setWidget(aliasWidget)
            self.addDockWidget(QtCore.Qt.RightDockWidgetArea, dock)
            self.controlDocks[alias.name] = dock
            self.cellControls[alias.name] = aliasWidget.value
            row += 1
            self.aliasWidgets[alias.name] = aliasWidget

        # Added a stretch space
        stretchDock = QCustomDockWidget('Stretch Space', self)
        stretch = QtGui.QWidget()
        stretch.setLayout(QtGui.QVBoxLayout())
        stretch.layout().addStretch()
        stretchDock.setWidget(stretch)
        self.addDockWidget(QtCore.Qt.RightDockWidgetArea, stretchDock)
        self.controlDocks["_stretch_"] = stretchDock

    def widget_changed(self, info):
        if self.cb_auto_update.isChecked() and not self.is_executing:
            self.updateCells(info)


    def run(self, useDefaultValues=False):

        # Building the list of parameter values
        params = []
        if useDefaultValues:
            for alias in self.currentMashup.alias_list:
                params.append((alias.component.vttype, alias.component.vtid,
                              alias.component.val))
        else:
            for (aliasName, edit) in self.cellControls.iteritems():
                alias = self.currentMashup.getAliasByName(aliasName)
                if hasattr(edit, 'contents'):
                    val = str(edit.contents())
                else:
                    val =str(edit.text())
                params.append((alias.component.vttype, alias.component.vtid,
                              val))
        results = self.controller.execute(params)[0]
        result = results[0]
        (objs, errors, executed) = (result.objects, result.errors,
                                                   result.executed)
        if len(errors) > 0:
            print '=== ERROR EXECUTING PIPELINE ==='
            print errors
            return (False, errors)
        return (True, [])

    def showBuilderWindow(self):
        from vistrails.gui.vistrails_window import _app
        _app.show()

class QCustomDockWidget(QtGui.QDockWidget):
    def __init__(self, title, parent=None):
        QtGui.QDockWidget.__init__(self, title, parent)
        self.setObjectName(title)
        self.setFeatures(QtGui.QDockWidget.DockWidgetClosable |
                         QtGui.QDockWidget.DockWidgetMovable)
        self.emptyTitleBar = QtGui.QWidget()
        self.titleBarVisible = True
        self.hideTitleBar()

    def showTitleBar(self):
        self.titleBarVisible = True
        self.setFeatures(QtGui.QDockWidget.DockWidgetClosable |
                         QtGui.QDockWidget.DockWidgetMovable)
        self.setMaximumHeight(524287)
        self.setTitleBarWidget(None)

    def hideTitleBar(self):
        self.titleBarVisible = False
        self.setFeatures(QtGui.QDockWidget.NoDockWidgetFeatures)
        self.setTitleBarWidget(self.emptyTitleBar)

    def toggleTitleBar(self):
        if self.titleBarVisible:
            self.hideTitleBar()
        else:
            self.showTitleBar()

################################################################################
# Testing


class TestMashupApp(TestVisTrailsGUI):
    def setUp(self):
        super(TestMashupApp, self).setUp()
        try:
            import vtk
        except ImportError:
            self.skipTest("VTK is not available")
        from vistrails.tests.utils import enable_package
        enable_package('org.vistrails.vistrails.vtk')

    def test_load_mashup(self):
        import vistrails.api
        import vistrails.core.system
        filename = (vistrails.core.system.vistrails_root_directory() +
                    '/tests/resources/spx_loop.vt')
        view = vistrails.api.open_vistrail_from_file(filename)
        # Execute workflow to trigger upgrades
        view.controller.execute_current_workflow()
        view.controller.flush_delayed_actions()
        id = "d5026457-de6c-11e2-b074-3c07543dba07"
        mashup = view.get_mashup_from_mashuptrail_id(id, "loop")
        self.assert_(mashup)
        view.open_mashup(mashup)
        mashup = view.get_mashup_from_mashuptrail_id(id, "no loop")
        self.assert_(mashup)
        view.open_mashup(mashup)



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
from __future__ import division

from vistrails.db.domain import DBControlParameter

import unittest
import copy

class ModuleControlParam(DBControlParameter):

    # Valid control parameters should be put here
    LOOP_KEY = 'loop_type' # How input lists are combined
    WHILE_COND_KEY = 'while_cond' # Run module in a while loop
    WHILE_INPUT_KEY = 'while_input' # input port for forwarded value
    WHILE_OUTPUT_KEY = 'while_output' # output port for forwarded value
    WHILE_MAX_KEY = 'while_max' # Max iterations
    WHILE_DELAY_KEY = 'while_delay' # delay between iterations
    CACHE_KEY = 'cache' # Turn caching on/off for this module (not implemented)
    JOB_CACHE_KEY = 'job_cache' # Always persist output values to disk

    ##########################################################################
    # Constructors and copy

    def __init__(self, *args, **kwargs):
        DBControlParameter.__init__(self, *args, **kwargs)
        if self.id is None:
            self.id = -1
        
    def __copy__(self):
        return ModuleControlParam.do_copy(self)

    def do_copy(self, new_ids=False, id_scope=None, id_remap=None):
        cp = DBControlParameter.do_copy(self, new_ids, id_scope, id_remap)
        cp.__class__ = ModuleControlParam
        return cp

    @staticmethod
    def convert(_control_parameter):
        _control_parameter.__class__ = ModuleControlParam

    ##########################################################################
    # Properties

    id = DBControlParameter.db_id
    name = DBControlParameter.db_name
    value = DBControlParameter.db_value

    ##########################################################################
    # Operators
    
    def __str__(self):
        """__str__() -> str - Returns a string representation of an ModuleControlParam
        object. 

        """
        rep = "<controlParameter id=%s name=%s value=%s</controlParameter>"
        return  rep % (str(self.id), str(self.name), str(self.value))

    def __eq__(self, other):
        """ __eq__(other: ModuleControlParam) -> boolean
        Returns True if self and other have the same attributes. Used by == 
        operator. 
        
        """
        if type(self) != type(other):
            return False
        if self.name != other.name:
            return False
        if self.value != other.value:
            return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)

################################################################################
# Unit tests


class TestModuleControlParam(unittest.TestCase):

    def create_control_parameter(self, id_scope=None):
        from vistrails.db.domain import IdScope

        if id_scope is None:
            id_scope = IdScope()
        control_parameter = ModuleControlParam(id=id_scope.getNewId(ModuleControlParam.vtType),
                                name='name %s',
                                value='some value %s')
        return control_parameter

    def test_copy(self):
        from vistrails.db.domain import IdScope
        id_scope = IdScope()

        a1 = self.create_control_parameter(id_scope)
        a2 = copy.copy(a1)
        self.assertEquals(a1, a2)
        self.assertEquals(a1.id, a2.id)
        a3 = a1.do_copy(True, id_scope, {})
        self.assertEquals(a1, a3)
        self.assertNotEquals(a1.id, a3.id)

    def test_serialization(self):
        import vistrails.core.db.io
        a1 = self.create_control_parameter()
        xml_str = vistrails.core.db.io.serialize(a1)
        a2 = vistrails.core.db.io.unserialize(xml_str, ModuleControlParam)
        self.assertEquals(a1, a2)
        self.assertEquals(a1.id, a2.id)

    def test_str(self):
        a1 = self.create_control_parameter()
        str(a1)
